#//
#//-----------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2013 NVIDIA Corporation
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//-----------------------------------------------------------------------------

from ..dap import uvm_set_before_get_dap
from .uvm_object import UVMObject
from .uvm_object_globals import *
from .uvm_misc import uvm_bitstream_to_string, uvm_integral_to_string
from .sv import sv
from ..macros import uvm_object_utils, uvm_warning, uvm_error
from .uvm_scope_stack import UVMScopeStack
from .uvm_globals import uvm_string_to_bits


"""
File: UVM Recorders

The `UVMRecorder` class serves two purposes:
 - Firstly, it is an abstract representation of a record within a
   `UVMTrStream`.
 - Secondly, it is a policy object for recording fields ~into~ that
   record within the ~stream~.

"""



class UVMRecorder(UVMObject):
    """
    CLASS: UVMRecorder

    Abstract class which defines the ~recorder~ API.
    """

    #   // Variable- m_ids_by_recorder
    #   // A dict of integers, indexed by `UVMRecorder`s.  This
    #   // provides a unique 'id' or 'handle' for each recorder, which can be
    #   // used to identify the recorder.
    #   //
    #   // By default, neither ~m_ids_by_recorder~ or ~m_recorders_by_id~ are
    #   // used.  Recorders are only placed in the arrays when the user
    #   // attempts to determine the id for a recorder.
    #   local static integer m_ids_by_recorder[uvm_recorder]
    m_ids_by_recorder = {}

    def __init__(self, name="uvm_recorder"):
        UVMObject.__init__(self, name)
        #  // Variable- m_stream_dap
        #  // Data access protected reference to the stream
        #  uvm_set_before_get_dap#(uvm_tr_stream) m_stream_dap
        self.m_stream_dap = uvm_set_before_get_dap("stream_dap")
        #  // Variable- m_warn_null_stream
        #  // Used to limit the number of warnings
        self.m_warn_null_stream = True
        #
        #  // Variable- m_is_opened
        #  // Used to indicate recorder is open
        self.m_is_opened = False
        #
        #  // Variable- m_is_closed
        #  // Used to indicate recorder is closed
        self.m_is_closed = False
        #
        #  // !m_is_opened && !m_is_closed == m_is_freed
        #
        #  // Variable- m_open_time
        #  // Used to store the open_time
        self.m_open_time = 0
        #
        #  // Variable- m_close_time
        #  // Used to store the close_time
        self.m_close_time = 0
        #
        #  // Variable- recording_depth
        self.recording_depth = 0
        #
        #  // Variable: default_radix
        #  //
        #  // This is the default radix setting if <record_field> is called without
        #  // a radix.
        #
        self.default_radix = UVM_HEX

        #
        #  // Variable: physical
        #  //
        #  // This bit provides a filtering mechanism for fields.
        #  //
        #  // The <abstract> and physical settings allow an object to distinguish between
        #  // two different classes of fields.
        #  //
        #  // It is up to you, in the <uvm_object::do_record> method, to test the
        #  // setting of this field if you want to use the physical trait as a filter.
        #
        self.physical = True
        #
        #
        #  // Variable: abstract
        #  //
        #  // This bit provides a filtering mechanism for fields.
        #  //
        #  // The abstract and physical settings allow an object to distinguish between
        #  // two different classes of fields.
        #  //
        #  // It is up to you, in the <uvm_object::do_record> method, to test the
        #  // setting of this field if you want to use the abstract trait as a filter.
        #
        self.abstract = True
        #
        #
        #  // Variable: identifier
        #  //
        #  // This bit is used to specify whether or not an object's reference should be
        #  // recorded when the object is recorded.
        #
        self.identifier = 1
        #
        #
        #  // Variable: recursion_policy
        #  //
        #  // Sets the recursion policy for recording objects.
        #  //
        #  // The default policy is deep (which means to recurse an object).
        ##  uvm_recursion_policy_enum
        self.policy = UVM_DEFAULT_POLICY
        #

    def get_stream(self):
        """
        Group: Configuration API

        Function: get_stream
        Returns a reference to the stream which created
        this record.

        A warning will be asserted if get_stream is called prior
        to the record being initialized via `do_open`.

        Returns:
        """
        astr = []
        if not self.m_stream_dap.try_get(astr):
            if self.m_warn_null_stream == 1:
               uvm_warning("UVM/REC/NO_CFG",
                   sv.sformatf("attempt to retrieve STREAM from '%s' before it was set!",
                       self.get_name()))
            self.m_warn_null_stream = 0
        if len(astr):
            return astr[0]


    #   // Group: Transaction Recorder API
    #   //
    #   // Once a recorder has been opened via <uvm_tr_stream::open_recorder>, the user
    #   // can ~close~ the recorder.
    #   //
    #   // Due to the fact that many database implementations will require crossing
    #   // a language boundary, an additional step of ~freeing~ the recorder is required.
    #   //
    #   // A ~link~ can be established within the database any time between ~open~ and
    #   // ~free~, however it is illegal to establish a link after ~freeing~ the recorder.
    #   //


    def close(self, close_time=0):
        """
           Function: close
           Closes this recorder.

           Closing a recorder marks the end of the transaction in the stream.

           Parameters:
           close_time - Optional time to record as the closing time of this transaction.

           This method will trigger a `do_close` call.
        Args:
            close_time:
        """
        if (close_time == 0):
            close_time = sv.realtime()

        if self.is_open() is False:
            return

        self.do_close(close_time)

        self.m_is_opened = 0
        self.m_is_closed = 1
        self.m_close_time = close_time


    def free(self, close_time=0):
        """
        Function: free
        Frees this recorder

        Freeing a recorder indicates that the stream and database can release
        any references to the recorder.

        If a recorder has not yet been closed (via a call to `close`), then
        `close` will automatically be called, and passed the `close_time`.  If the recorder
        has already been closed, then the `close_time` will be ignored.

        This method will trigger a `do_free` call.

        Args:
            close_time (int): Optional time to record as the closing time of this transaction.
        """
        # process p=process::self()
        p = None
        s = ""
        stream = None  # uvm_tr_stream

        if not self.is_open() and not self.is_closed():
            return

        if self.is_open():
            self.close(close_time)

        self.do_free()

        # Clear out internal state
        stream = self.get_stream()

        self.m_is_closed = 0
        if p is not None:
            s = p.get_randstate()
        self.m_stream_dap = uvm_set_before_get_dap("stream_dap")
        if p is not None:
            p.set_randstate(s)
        self.m_warn_null_stream = 1
        if self in UVMRecorder.m_ids_by_recorder:
            UVMRecorder.m_free_id(UVMRecorder.m_ids_by_recorder[self])

        # Clear out stream state
        if stream is not None:
            stream.m_free_recorder(self)

    def is_open(self):
        """
           Function: is_open
           Returns true if this `uvm_recorder` was opened on its stream,
           but has not yet been closed.

        Returns:
        """
        return self.m_is_opened


    def get_open_time(self):
        """

           Function: get_open_time
           Returns the `open_time`

        Returns:
        """
        return self.m_open_time


    def is_closed(self):
        """
           Function: is_closed
           Returns true if this `uvm_recorder` was closed on its stream,
           but has not yet been freed.

        Returns:
        """
        return self.m_is_closed


    def get_close_time(self):
        """
           Function: get_close_time
           Returns the `close_time`

        Returns:
        """
        return self.m_close_time

    def m_do_open(self, stream, open_time, type_name):
        """
          Function- m_do_open
          Initializes the internal state of the recorder.

          Parameters:
          stream - The stream which spawned this recorder

          This method will trigger a `do_open` call.

          An error will be asserted if:
          - `m_do_open` is called more than once without the
           recorder being `freed` in between.
          - `stream` is `null`
        Args:
            stream:
            open_time:
            type_name:
        """
        m_stream = []
        if stream is None:
            uvm_error("UVM/REC/NULL_STREAM",
               sv.sformatf("Illegal attempt to set STREAM for '%s' to '<null>'",
                 self.get_name()))
            return

        if self.m_stream_dap.try_get(m_stream):
            uvm_error("UVM/REC/RE_INIT",
                sv.sformatf("Illegal attempt to re-initialize '%s'",
                self.get_name()))
            return

        self.m_stream_dap.set(stream)
        self.m_open_time = open_time
        self.m_is_opened = 1

        self.do_open(stream, open_time, type_name)


    #   // Group: Handles

    #   // Variable- m_recorders_by_id
    #   // A corollary to ~m_ids_by_recorder~, this indexes the recorders by their
    #   // unique ids.
    #   local static uvm_recorder m_recorders_by_id[integer]
    m_recorders_by_id = {}


    #   // Variable- m_id
    #   // Static int marking the last assigned id.
    #   local static integer m_id


    @classmethod
    def m_free_id(cls, id):
        """
           Function- m_free_id
           Frees the id/recorder link (memory cleanup)
        Args:
            cls:
            id:
        """
        recorder = None  # uvm_recorder
        if ((not sv.isunknown(id)) and (id in cls.m_recorders_by_id)):
            recorder = cls.m_recorders_by_id[id]

        if recorder is not None:
            del cls.m_recorders_by_id[id]
            del cls.m_ids_by_recorder[recorder]

    def get_handle(self):
        """
           Function: get_handle
           Returns a unique ID for this recorder.

           A value of `0` indicates that the recorder has been `freed`,
           and no longer has a valid ID.

        Returns:
        """
        if (self.is_open() is False and self.is_closed() is False):
            return 0
        else:
            handle = self.get_inst_id()

            m_ids_by_rec = UVMRecorder.m_ids_by_recorder
            # Check for the weird case where our handle changed.
            if self in m_ids_by_rec and m_ids_by_rec[self] != handle:
                this_id = UVMRecorder.m_ids_by_recorder[self]
                del UVMRecorder.m_recorders_by_id[this_id]

            UVMRecorder.m_recorders_by_id[handle] = self
            UVMRecorder.m_ids_by_recorder[self] = handle
            return handle
        #   endfunction : get_handle


    @classmethod
    def get_recorder_from_handle(cls, id):
        """
        Static accessor, returns a recorder reference for a given unique id.

        If no recorder exists with the given `id`, or if the
        recorder with that `id` has been freed, then `null` is
        returned.

        This method can be used to access the recorder associated with a
        call to `UVMTransaction.begin_tr` or `UVMComponent.begin_tr`.

        .. code-block:: python

           handle = tr.begin_tr()
           recorder = UVMRecorder.get_recorder_from_handle(handle)
           if recorder is not None:
               recorder.record_string("begin_msg", "Started recording transaction!")

        Args:
            cls:
            id:
        Returns:
            UVMRecorder:
        """
        if id == 0:
            return None
        if sv.isunknown(id) or id not in cls.m_recorders_by_id:
            return None
        return cls.m_recorders_by_id[id]


    #   // Group: Attribute Recording
    #
    #   // Function: record_field
    #   // Records an integral field (less than or equal to 4096 bits).
    #   //
    #   // Parameters:
    #   // name - Name of the field
    #   // value - Value of the field to record.
    #   // size - Number of bits of the field which apply (Usually obtained via $bits).
    #   // radix - The <uvm_radix_enum> to use.
    #   //
    #   // This method will trigger a <do_record_field> call.
    def record_field(self, name, value, size, radix=UVM_NORADIX):
        if self.get_stream() is None:
            return
        self.do_record_field(name, value, size, radix)


    #   // Function: record_field_int
    #   // Records an integral field (less than or equal to 64 bits).
    #   //
    #   // This optimized version of <record_field> is useful for sizes up
    #   // to 64 bits.
    #   //
    #   // Parameters:
    #   // name - Name of the field
    #   // value - Value of the field to record
    #   // size - Number of bits of the wfield which apply (Usually obtained via $bits).
    #   // radix - The <uvm_radix_enum> to use.
    #   //
    #   // This method will trigger a <do_record_field_int> call.
    def record_field_int(self,name, value, size, radix=UVM_NORADIX):
        if self.get_stream() is None:
            return
        self.do_record_field_int(name, value, size, radix)


    #   // Function: record_field_real
    #   // Records a real field.
    #   //
    #   // Parameters:
    #   // name - Name of the field
    #   // value - Value of the field to record
    #   //
    #   // This method will trigger a <do_record_field_real> call.
    #   function void record_field_real(string name,
    #                                   real value)
    #      if (get_stream() is None):
    #         return
    #      end
    #      do_record_field_real(name, value)
    #   endfunction : record_field_real


    #   // Function: record_object
    #   // Records an object field.
    #   //
    #   // Parameters:
    #   // name - Name of the field
    #   // value - Object to record
    #   //
    #   // The implementation must use the <recursion_policy> and <identifier> to
    #   // determine exactly what should be recorded.
    def record_object(self, name, value):
        if (self.get_stream() is None):
            return
        self.do_record_object(name, value)


    #   // Function: record_string
    #   // Records a string field.
    #   //
    #   // Parameters:
    #   // name - Name of the field
    #   // value - Value of the field
    #   //
    def record_string(self, name, value):
        if (self.get_stream() is None):
            return
        self.do_record_string(name, value)


    #   // Function: record_time
    #   // Records a time field.
    #   //
    #   // Parameters:
    #   // name - Name of the field
    #   // value - Value of the field
    #   //
    #   function void record_time(string name,
    #                             time value)
    #      if (get_stream() is None):
    #         return
    #      end
    #
    #      do_record_time(name, value)
    #   endfunction : record_time
    #
    #   // Function: record_generic
    #   // Records a name/value pair, where ~value~ has been converted to a string.
    #   //
    #   // For example:
    #   //| recorder.record_generic("myvar","var_type", sv.sformatf("%0d",myvar), 32)
    #   //
    #   // Parameters:
    #   // name - Name of the field
    #   // value - Value of the field
    #   // type_name - ~optional~ Type name of the field
    #   function void record_generic(string name,
    #                                string value,
    #                                string type_name="")
    #      if (get_stream() is None):
    #         return
    #      end
    #
    #      do_record_generic(name, value, type_name)
    #   endfunction : record_generic


    #  // Function: use_record_attribute
    #  //
    #  // Indicates that this recorder does (or does not) support usage of
    #  // the <`uvm_record_attribute> macro.
    #  //
    #  // The default return value is ~0~ (not supported), developers can
    #  // optionally extend ~uvm_recorder~ and set the value to ~1~ if they
    #  // support the <`uvm_record_attribute> macro.
    def use_record_attribute(self):
        return 0


    #   // Function: get_record_attribute_handle
    #   // Provides a tool-specific handle which is compatible with <`uvm_record_attribute>.
    #   //
    #   // By default, this method will return the same value as <get_handle>,
    #   // however tool vendors can override this method to provide tool-specific handles
    #   // which will be passed to the <`uvm_record_attribute> macro.
    #   //
    #   virtual function integer get_record_attribute_handle()
    #      return get_handle()
    #   endfunction : get_record_attribute_handle
    #
    #   // Group: Implementation Agnostic API
    #

    def do_open(self, stream, open_time, type_name):
        """
           Function: do_open
           Callback triggered via <uvm_tr_stream::open_recorder>.

           The `do_open` callback can be used to initialize any internal
           state within the recorder, as well as providing a location to
           record any initial information.
        Args:
            stream:
            open_time:
            type_name:
        """
        pass

    #
    #   // Function: do_close
    #   // Callback triggered via <close>.
    #   //
    #   // The ~do_close~ callback can be used to set internal state
    #   // within the recorder, as well as providing a location to
    #   // record any closing information.
    def do_close(self, close_time):
        pass

    #   // Function: do_free
    #   // Callback triggered via <free>.
    #   //
    #   // The ~do_free~ callback can be used to release the internal
    #   // state within the recorder, as well as providing a location
    #   // to record any "freeing" information.
    def do_free(self):
        pass


    #   // Function: do_record_field
    #   // Records an integral field (less than or equal to 4096 bits).
    #   //
    #   // ~Mandatory~ Backend implementation of <record_field>
    #   pure virtual protected function void do_record_field(string name,
    #                                                        uvm_bitstream_t value,
    #                                                        int size,
    #                                                        uvm_radix_enum radix)


    #   // Function: do_record_field_int
    #   // Records an integral field (less than or equal to 64 bits).
    #   //
    #   // ~Mandatory~ Backend implementation of <record_field_int>
    #   pure virtual protected function void do_record_field_int(string name,
    #                                                            uvm_integral_t value,
    #                                                            int          size,
    #                                                            uvm_radix_enum radix)
    #
    #   // Function: do_record_field_real
    #   // Records a real field.
    #   //
    #   // ~Mandatory~ Backend implementation of <record_field_real>
    #   pure virtual protected function void do_record_field_real(string name,
    #                                                             real value)
    #
    #   // Function: do_record_object
    #   // Records an object field.
    #   //
    #   // ~Mandatory~ Backend implementation of <record_object>
    #   pure virtual protected function void do_record_object(string name,
    #                                                         uvm_object value)
    #
    #   // Function: do_record_string
    #   // Records a string field.
    #   //
    #   // ~Mandatory~ Backend implementation of <record_string>
    #   pure virtual protected function void do_record_string(string name,
    #                                                         string value)
    #
    #   // Function: do_record_time
    #   // Records a time field.
    #   //
    #   // ~Mandatory~ Backend implementation of <record_time>
    #   pure virtual protected function void do_record_time(string name,
    #                                                       time value)
    #
    #   // Function: do_record_generic
    #   // Records a name/value pair, where ~value~ has been converted to a string.
    #   //
    #   // ~Mandatory~ Backend implementation of <record_generic>
    #   pure virtual protected function void do_record_generic(string name,
    #                                                          string value,
    #                                                          string type_name)
    #
    #
    #   // The following code is primarily for backwards compat. purposes.  "Transaction
    #   // Handles" are useful when connecting to a backend, but when passing the information
    #   // back and forth within simulation, it is safer to user the ~recorder~ itself
    #   // as a reference to the transaction within the database.
    #
    #   //------------------------------
    #   // Group- Vendor-Independent API
    #   //------------------------------
    #
    #
    #  // UVM provides only a text-based default implementation.
    #  // Vendors provide subtype implementations and overwrite the
    #  // <uvm_default_recorder> handle.
    #
    #
    #  // Function- open_file
    #  //
    #  // Opens the file in the <filename> property and assigns to the
    #  // file descriptor <file>.
    #  //
    #  virtual function bit open_file()
    #     return 0
    #  endfunction
    #
    #  // Function- create_stream
    #  //
    #  //
    #  virtual function integer create_stream (string name,
    #                                          string t,
    #                                          string scope)
    #     return -1
    #  endfunction
    #
    #
    #  // Function- m_set_attribute
    #  //
    #  //
    #  virtual function void m_set_attribute (integer txh,
    #                                 string nm,
    #                                 string value)
    #  endfunction
    #
    #
    #  // Function- set_attribute
    #  //
    #  virtual function void set_attribute (integer txh,
    #                               string nm,
    #                               logic [1023:0] value,
    #                               uvm_radix_enum radix,
    #                               integer numbits=1024)
    #  endfunction
    #
    #
    #  // Function- check_handle_kind
    #  //
    #  //
    #  virtual function integer check_handle_kind (string htype, integer handle)
    #     return 0
    #  endfunction
    #
    #
    #  // Function- begin_tr
    #  //
    #  //
    #  virtual function integer begin_tr(string txtype,
    #                                     integer stream,
    #                                     string nm,
    #                                     string label="",
    #                                     string desc="",
    #                                     time begin_time=0)
    #    return -1
    #  endfunction
    #
    #
    #  // Function- end_tr
    #  //
    #  //
    #  virtual function void end_tr (integer handle, time end_time=0)
    #  endfunction
    #
    #
    #  // Function- link_tr
    #  //
    #  //
    #  virtual function void link_tr(integer h1,
    #                                 integer h2,
    #                                 string relation="")
    #  endfunction
    #
    #
    #
    #  // Function- free_tr
    #  //
    #  //
    #  virtual function void free_tr(integer handle)
    #  endfunction
    #
    #endclass // uvm_recorder
#
#//------------------------------------------------------------------------------
#//
#// CLASS: uvm_text_recorder
#//
#// The ~uvm_text_recorder~ is the default recorder implementation for the
#// <uvm_text_tr_database>.
#//

class UVMTextRecorder(UVMRecorder):

    def __init__(self, name="unnamed-uvm_text_recorder"):
        """
           Function: new
           Constructor

           Parameters:
           name - Instance name
        Args:
            name:
        """
        super().__init__(name)
        # Variable- m_text_db
        # Reference to the text database backend
        self.m_text_db = None  # #   uvm_text_tr_database

        # Variable- scope
        # Imeplementation detail
        self.scope = UVMScopeStack()


    def do_open(self, stream, open_time, type_name):
        """
           Group: Implementation Agnostic API

           Function: do_open
           Callback triggered via <uvm_tr_stream::open_recorder>.

           Text-backend specific implementation.
        Args:
            stream:
            open_time:
            type_name:
        """
        self. m_text_db = stream.get_db()
        if self.m_text_db.open_db():
            sv.fdisplay(self.m_text_db.m_file,
                "    OPEN_RECORDER @%0t {{TXH:%0d STREAM:%0d NAME:%s TIME:%0t TYPE=\"%0s\"}}",
                sv.realtime(), self.get_handle(), stream.get_handle(),
                self.get_name(), open_time, type_name)


    def do_close(self, close_time):
        """
           Function: do_close
           Callback triggered via <uvm_recorder::close>.

           Text-backend specific implementation.
        Args:
            close_time:
        """
        if self.m_text_db.open_db():
            sv.fdisplay(self.m_text_db.m_file,
                 "    CLOSE_RECORDER @%0t {{TXH:%0d TIME=%0t}}",
                 sv.realtime(),
                 self.get_handle(),
                 close_time)

    def do_free(self):
        """
        Function: do_free
        Callback triggered via <uvm_recorder::free>.

        Text-backend specific implementation.
        """
        if self.m_text_db.open_db():
            sv.fdisplay(self.m_text_db.m_file,
                "    FREE_RECORDER @%0t {{TXH:%0d}}",
                sv.realtime(), self.get_handle())

        self.m_text_db = None


    #   // Function: do_record_field
    #   // Records an integral field (less than or equal to 4096 bits).
    #   //
    #   // Text-backend specific implementation.
    #   protected virtual function void do_record_field(string name,
    #                                                   uvm_bitstream_t value,
    #                                                   int size,
    #                                                   uvm_radix_enum radix)
    def do_record_field(self, name, value, size, radix):
        self.scope.set_arg(name)
        if not radix:
            radix = self.default_radix
        self.write_attribute(self.scope.get(), value, radix, size)


    #   // Function: do_record_field_int
    #   // Records an integral field (less than or equal to 64 bits).
    #   //
    #   // Text-backend specific implementation.
    def do_record_field_int(self, name, value, size, radix):
        self.scope.set_arg(name)
        if not radix:
            radix = self.default_radix
        self.write_attribute_int(self.scope.get(), value, radix, size)


    #   // Function: do_record_field_real
    #   // Record a real field.
    #   //
    #   // Text-backened specific implementation.
    #   protected virtual function void do_record_field_real(string name,
    #                                                        real value)
    #      bit [63:0] ival = $realtobits(value)
    #      scope.set_arg(name)
    #
    #      write_attribute_int(scope.get(),
    #                          ival,
    #                          UVM_REAL,
    #                          64)
    #   endfunction : do_record_field_real
    #


    #   // Function: do_record_object
    #   // Record an object field.
    #   //
    #   // Text-backend specific implementation.
    #   //
    #   // The method uses ~identifier~ to determine whether or not to
    #   // record the object instance id, and ~recursion_policy~ to
    #   // determine whether or not to recurse into the object.
    def do_record_object(self, name, value):
        v = 0
        _str = ""

        if self.identifier:
            if value is not None:
                _str = str(value.get_inst_id())
                v = int(_str)
            self.scope.set_arg(name)
            self.write_attribute_int(self.scope.get(), v, UVM_DEC, 32)

        if self.policy != UVM_REFERENCE:
            if value is not None:
                if value in value._m_uvm_status_container.cycle_check:
                    return
                value._m_uvm_status_container.cycle_check[value] = 1
                self.scope.down(name)
                value.record(self)
                self.scope.up()
                del value._m_uvm_status_container.cycle_check[value]


    #   // Function: do_record_string
    #   // Records a string field.
    #   //
    #   // Text-backend specific implementation.
    def do_record_string(self, name, value):
        self.scope.set_arg(name)
        if self.m_text_db.open_db():
            sv.fdisplay(self.m_text_db.m_file,
                "      SET_ATTR @%0t {{TXH:%0d NAME:%s VALUE:%s   RADIX:%s BITS=%0d}}",
                sv.realtime(), self.get_handle(), self.scope.get(),
                value, "UVM_STRING", 8 + len(value))


    #   // Function: do_record_time
    #   // Records a time field.
    #   //
    #   // Text-backend specific implementation.
    def do_record_time(self, name,value):
        self.scope.set_arg(name)
        self.write_attribute_int(self.scope.get(), value, UVM_TIME, 64)


    #   // Function: do_record_generic
    #   // Records a name/value pair, where ~value~ has been converted to a string.
    #   //
    #   // Text-backend specific implementation.
    def do_record_generic(self, name, value, type_name):
        self.scope.set_arg(name)
        self.write_attribute(self.scope.get(), uvm_string_to_bits(value), UVM_STRING,
            8+len(value))


    #   // Group: Implementation Specific API


    #   // Function: write_attribute
    #   // Outputs an integral attribute to the textual log
    #   //
    #   // Parameters:
    #   // nm - Name of the attribute
    #   // value - Value
    #   // radix - Radix of the output
    #   // numbits - number of valid bits
    #   function void write_attribute(string nm,
    #                                 uvm_bitstream_t value,
    #                                 uvm_radix_enum radix,
    #                                 integer numbits=$bits(uvm_bitstream_t))
    def write_attribute(self, nm, value, radix, numbits=32):
        if self.m_text_db.open_db():
            sv.fdisplay(self.m_text_db.m_file,
                "      SET_ATTR @%0t {{TXH:%0d NAME:%s VALUE:%s   RADIX:%s BITS=%0d}}",
                 sv.realtime(),
                 self.get_handle(),
                 nm,
                 uvm_bitstream_to_string(value, numbits, radix),
                 UVM_RADIX_TO_STRING_DICT[radix],
                 numbits)


    #   // Function: write_attribute_int
    #   // Outputs an integral attribute to the textual log
    #   //
    #   // Parameters:
    #   // nm - Name of the attribute
    #   // value - Value
    #   // radix - Radix of the output
    #   // numbits - number of valid bits
    def write_attribute_int(self, nm, value, radix, numbits=32):
        if self.m_text_db.open_db():
            sv.fdisplay(self.m_text_db.m_file,
                "      SET_ATTR @%0t {{TXH:%0d NAME:%s VALUE:%s   RADIX:%s BITS=%0d}}",
                sv.realtime(), self.get_handle(), nm,
                uvm_integral_to_string(value, numbits, radix),
                UVM_RADIX_TO_STRING_DICT[radix], numbits)


    #  // UVM provides only a text-based default implementation.
    #  // Vendors provide subtype implementations and overwrite the
    #  // <uvm_default_recorder> handle.
    #
    #   string                                                   filename
    #   bit                                                      filename_set
    #
    #  // Function- open_file
    #  //
    #  // Opens the file in the <filename> property and assigns to the
    #  // file descriptor <file>.
    #  //
    #  virtual function bit open_file()
    #     if (!filename_set):
    #        m_text_db.set_file_name(filename)
    #     end
    #     return m_text_db.open_db()
    #  endfunction


    #  // Function- create_stream
    #  //
    #  //
    #  virtual function integer create_stream (string name,
    #                                          string t,
    #                                          string scope)
    #     uvm_text_tr_stream stream
    #     if (open_file()):
    #        $cast(stream,m_text_db.open_stream(name, scope, t))
    #        return stream.get_handle()
    #     end
    #     return 0
    #  endfunction


    #  // Function- m_set_attribute
    #  //
    #  //
    #  virtual function void m_set_attribute (integer txh,
    #                                 string nm,
    #                                 string value)
    #     if (open_file()):
    #        UVM_FILE file = m_text_db.m_file
    #        $fdisplay(file,"      SET_ATTR @%0t {TXH:%0d NAME:%s VALUE:%s}", $realtime,txh,nm,value)
    #     end
    #  endfunction

    #
    #
    #  // Function- set_attribute
    #  //
    #  //
    #  virtual function void set_attribute (integer txh,
    #                               string nm,
    #                               logic [1023:0] value,
    #                               uvm_radix_enum radix,
    #                               integer numbits=1024)
    #     if (open_file()):
    #        UVM_FILE file = m_text_db.m_file
    #         $fdisplay(file,
    #                   "      SET_ATTR @%0t {TXH:%0d NAME:%s VALUE:%s   RADIX:%s BITS=%0d}",
    #                   $realtime,
    #                   txh,
    #                   nm,
    #                   uvm_bitstream_to_string(value, numbits, radix),
    #                   radix.name(),
    #                   numbits)
    #
    #     end
    #  endfunction

    #
    #
    #  // Function- check_handle_kind
    #  //
    #  //
    #  virtual function integer check_handle_kind (string htype, integer handle)
    #     return ((uvm_recorder::get_recorder_from_handle(handle) != null) ||
    #             (uvm_tr_stream::get_stream_from_handle(handle) != null))
    #  endfunction


    #  // Function- begin_tr
    #  //
    #  //
    #  virtual function integer begin_tr(string txtype,
    #                                     integer stream,
    #                                     string nm,
    #                                     string label="",
    #                                     string desc="",
    #                                     time begin_time=0)
    #     if (open_file()):
    #        uvm_tr_stream stream_obj = uvm_tr_stream::get_stream_from_handle(stream)
    #        uvm_recorder recorder
    #
    #        if (stream_obj is None)
    #          return -1
    #
    #        recorder = stream_obj.open_recorder(nm, begin_time, txtype)
    #
    #        return recorder.get_handle()
    #     end
    #     return -1
    #  endfunction

    #
    #
    #  // Function- end_tr
    #  //
    #  //
    #  virtual function void end_tr (integer handle, time end_time=0)
    #     if (open_file()):
    #        uvm_recorder record = uvm_recorder::get_recorder_from_handle(handle)
    #        if (record != null):
    #           record.close(end_time)
    #        end
    #     end
    #  endfunction

    #
    #
    #  // Function- link_tr
    #  //
    #  //
    #  virtual function void link_tr(integer h1,
    #                                 integer h2,
    #                                 string relation="")
    #    if (open_file())
    #      $fdisplay(self.m_text_db.m_file,"  LINK @%0t {TXH1:%0d TXH2:%0d RELATION=%0s}", $realtime,h1,
    #          h2,relation)
    #  endfunction


    #  // Function- free_tr
    #  //
    #  //
    #  virtual function void free_tr(integer handle)
    #     if (open_file()):
    #        uvm_recorder record = uvm_recorder::get_recorder_from_handle(handle)
    #        if (record != null):
    #           record.free()
    #        end
    #     end
    #  endfunction // free_tr


uvm_object_utils(UVMTextRecorder)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010-2011 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------

import cocotb
from cocotb.triggers import *

from uvm.base.uvm_callback import *
from uvm.base.uvm_config_db import *
from uvm.comps.uvm_driver import UVMDriver
from uvm.macros import *

from .apb_rw import *


class apb_master_cbs(UVMCallback):
    
    async def trans_received(self, xactor, cycle):
        await Timer(1, "NS")

    
    async def trans_executed(self, xactor, cycle):
        await Timer(1, "NS")


class apb_master(UVMDriver):  #(apb_rw)


    #   event trig
    #   apb_vif sigs
    #   apb_config cfg
    #
    def __init__(self, name, parent=None):
        super().__init__(name,parent)
        self.trig = Event("trans_exec")  # event
        self.sigs = None  # apb_vif
        self.cfg = None  # apb_config
        self.tag = "APB_MASTER"


    def build_phase(self, phase):
        super().build_phase(phase)
        agent = self.get_parent()
        if agent is not None:
            self.sigs = agent.vif
        else:
            arr = []
            if (not UVMConfigDb.get(self, "", "vif", arr)):
                uvm_fatal("APB/DRV/NOVIF", "No virtual interface specified for self driver instance")
            else:
                self.sigs = arr[0]


    
    async def run_phase(self, phase):
        uvm_info("APB_MASTER", "apb_master run_phase started", UVM_MEDIUM)

        self.sigs.psel    <= 0
        self.sigs.penable <= 0

        while True:
            await self.drive_delay()

            tr = []
            await self.seq_item_port.get_next_item(tr)
            tr = tr[0]
            uvm_info("APB_MASTER", "Driving trans into DUT: " + tr.convert2string(), UVM_DEBUG)

            #if (not self.sigs.clk.triggered):
            #yield Edge(self.sigs.clk)
            await self.drive_delay()
            #yield RisingEdge(self.sigs.clk)
            #yield Timer(1, "NS")

            await self.trans_received(tr)
            #uvm_do_callbacks(apb_master,apb_master_cbs,trans_received(self,tr))

            if tr.kind == apb_rw.READ:
                data = []
                await self.read(tr.addr, data)
                tr.data = data[0]
            elif tr.kind == apb_rw.WRITE:
                await self.write(tr.addr, tr.data)

            await self.trans_executed(tr)
            #uvm_do_callbacks(apb_master,apb_master_cbs,trans_executed(self,tr))

            self.seq_item_port.item_done()
            self.trig.set()
            #->trig
        #   endtask: run_phase


    
    async def read(self, addr, data):
        uvm_info(self.tag, "Doing APB read to addr " + str(addr), UVM_MEDIUM)

        self.sigs.paddr   <= addr
        self.sigs.pwrite  <= 0
        self.sigs.psel    <= 1
        await self.drive_delay()
        self.sigs.penable <= 1
        await self.drive_delay()
        data.append(int(self.sigs.prdata))
        self.sigs.psel    <= 0
        self.sigs.penable <= 0
        #   endtask: read


    
    async def write(self, addr, data):
        uvm_info(self.tag, "Doing APB write to addr " + str(addr), UVM_MEDIUM)
        self.sigs.paddr   <= addr
        self.sigs.pwdata  <= data
        self.sigs.pwrite  <= 1
        self.sigs.psel    <= 1
        await self.drive_delay()
        self.sigs.penable <= 1
        await self.drive_delay()
        self.sigs.psel    <= 0
        self.sigs.penable <= 0
        uvm_info(self.tag, "Finished APB write to addr " + str(addr), UVM_MEDIUM)
        #   endtask: write


    
    async def drive_delay(self):
        await RisingEdge(self.sigs.clk)
        await Timer(1, "NS")

    
    async def trans_received(self, tr):
        await Timer(1, "NS")

    
    async def trans_executed(self, tr):
        await Timer(1, "NS")


uvm_component_utils(apb_master)



#//
#// -------------------------------------------------------------
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//


from uvm.reg.uvm_reg_fifo import UVMRegFIFO
from uvm.macros import uvm_object_utils
from uvm.reg.uvm_reg_block import UVMRegBlock
from uvm.reg.uvm_reg_model import UVM_NO_COVERAGE, UVM_LITTLE_ENDIAN

FIFO_SIZE = 8


class fifo_reg(UVMRegFIFO):

    def __init__(self, name="fifo_reg"):
        super().__init__(name, FIFO_SIZE, 32, UVM_NO_COVERAGE)


uvm_object_utils(fifo_reg)


class reg_block_B(UVMRegBlock):

    #    rand fifo_reg FIFO

    def __init__(self, name="B"):
        super().__init__(name, UVM_NO_COVERAGE)


    def build(self):
        self.FIFO = fifo_reg.type_id.create("FIFO")
        self.FIFO.configure(self, None)
        self.FIFO.build()

        self.default_map = self.create_map("default_map_fifo", 0x0, 4, UVM_LITTLE_ENDIAN)
        self.default_map.add_reg(self.FIFO, 0x0, "RW")


uvm_object_utils(reg_block_B)



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer, Combine
from cocotb.utils import get_sim_time

from uvm.base.uvm_debug import UVMDebug
from uvm.base.sv import sv
from uvm.base.uvm_globals import run_test
from uvm.base.uvm_object_globals import UVM_MEDIUM, UVM_HIGH
from uvm.base.uvm_phase import UVMPhase
from uvm.seq import UVMSequence, UVMSequenceItem, UVMSequencer
from uvm.macros import *
from uvm.comps import UVMDriver, UVMEnv

NUM_SEQS = 1
NUM_LOOPS = 10
USE_FIELD_MACROS = 0

#UVMDebug.full_debug()

def toSTRING(X):
    if USE_FIELD_MACROS:
        return X.sprint()
    else:
        return X.convert2string()

#Remark 1:
#
#- this example shows two alternative approaches to handle local properties -
#
#When the define USE_FIELD_MACROS is set, the code is utilizing the uvm_field_xxx macros
#to mark transaction fields as "uvm fields". Doing so automatically infers code which handles
#compare,print,copy,record,pack,unpack without any additional implementation from the user side.
#This is typically shorter to write and easier to maintain on the user side. On the other hand a generic
#implementation might not be as fast and custom tailored.
#
#The two styles can be mixed for most of the uvm_field functionality (manually handling few fields and
#use a default implementation for some others). Only the component level auto configuration for fields is for
#`uvm_field_* annotated members only. it means that only when the uvm_field_* macros are used these fields  will be
#set to their configured values automatically.


#Remark 2:
#
#All classes in this example utilize the uvm_(object|component)[_param]_utils macro. Althrough its not mandatory
#for UVM it is good practice since this macro will register the class in the UVM factory automatically. See the user
#guide for more information regarding uvm factory use models


#bus_op_t = sv.enum(BUS_READ, BUS_WRITE)
#status_t = sv.enum(STATUS_OK, STATUS_NOT_OK)
BUS_READ = 0
BUS_WRITE = 1

STATUS_OK = 0
STATUS_NOT_OK = 1

class bus_trans(UVMSequenceItem):

    def __init__(self, name=""):
        UVMSequenceItem.__init__(self, name)
        self.addr = 0
        self.data = 0
        self.op = 0

    if USE_FIELD_MACROS == 0:
        def do_copy(self, rhs):
            #bus_trans rhs_
            rhs_ = rhs
            if not sv.cast(rhs_, rhs):
                uvm_error("do_copy", "cast failed, check type compatability")
            UVMSequenceItem.do_copy(self, rhs)

            self.addr = rhs_.addr
            self.data = rhs_.data
            self.op = rhs_.op

        def do_compare(self, rhs, comparer):
            rhs_ = rhs  # bus_trans rhs_
            #if(!$cast(rhs_, rhs))
            #        `uvm_fatal("do_compare", "cast failed, check type compatability")
            return (self.op == rhs_.op) and (self.addr == rhs_.addr) and (self.data == rhs_.data)

        def convert2string(self):
            return sv.sformatf("op %s: addr=%03x, data=%02x",
                str(self.op), self.addr, self.data)

        # NOTE: in contrast to the USE_FIELD_MACROS version this doesnt implement pack/unpack/print/record/...
        #`endif

if USE_FIELD_MACROS:
    uvm_object_utils_begin(bus_trans)
    uvm_field_int('addr',UVM_DEFAULT)
    uvm_field_int('data',UVM_DEFAULT)
    uvm_field_enum('op',UVM_DEFAULT)
    uvm_field_utils_end(bus_trans)
else:
    uvm_object_utils(bus_trans)

#//--------------------------------------------------------------------
#// bus_req
#//--------------------------------------------------------------------
class bus_req(bus_trans):
    def __init__(self, name=""):
        UVMSequenceItem.__init__(self, name)
uvm_object_utils(bus_req)

#//--------------------------------------------------------------------
#// bus_rsp
#//--------------------------------------------------------------------
class bus_rsp(bus_trans):

    def __init__(self, name=""):
        bus_trans.__init__(self, name)
        self.status = 0

    if USE_FIELD_MACROS:
        pass
    else:

        def do_copy(self, rhs):
            rhs_ = rhs
            #bus_rsp rhs_
            #if(!$cast(rhs_, rhs))
            #    uvm_fatal("do_copy", "cast failed, check type compatability")

            super.do_copy(rhs_)
            self.status = rhs_.status

        def convert2string(self):
            return sv.sformatf("op %s, status=%s",
                    bus_trans.convert2string(self), str(self.status))

        # NOTE: in contrast to the USE_FIELD_MACROS version this doesnt implement pack/unpack/print/record/...


if USE_FIELD_MACROS:
    uvm_object_utils_begin(bus_rsp)
    uvm_field_enum('status', UVM_DEFAULT)
    uvm_object_utils_end(bus_rsp)
else:
    uvm_object_utils(bus_rsp)


class my_driver(UVMDriver):

    def __init__(self, name, parent):
        UVMDriver.__init__(self, name, parent)
        self.data_array = [0] * 512  # int data_array[511:0]
        self.nitems = 0
        self.recording_detail = UVM_HIGH

    
    async def run_phase(self, phase):
        req = None
        rsp = None

        while True:
            uvm_info("DRIVER", "my_driver getting the req now..",
                UVM_MEDIUM)
            qreq = []
            await self.seq_item_port.get(qreq)

            self.nitems += 1
            req = qreq[0]
            self.begin_tr(req)
            rsp = bus_rsp("bus_rsp")
            rsp.set_id_info(req)

            # Actually do the read or write here
            if req.op == BUS_READ:
                rsp.addr = req.addr  # [8:0]
                rsp.data = self.data_array[rsp.addr]
                uvm_info("sending", toSTRING(rsp), UVM_MEDIUM)
            else:
                self.data_array[req.addr] = req.data
                uvm_info("sending", toSTRING(req), UVM_MEDIUM)
            await self.seq_item_port.put(rsp)
            self.end_tr(req)

uvm_component_utils(my_driver)


class sequenceA(UVMSequence):

    #local static integer g_my_id = 1
    g_my_id = 1

    def __init__(self, name=""):
        UVMSequence.__init__(self, name)
        self.my_id = sequenceA.g_my_id
        sequenceA.g_my_id += 1

    
    async def body(self):
        req = None  # REQ  req
        rsp = None  # RSP  rsp
        uvm_info("sequenceA", "Starting sequence", UVM_MEDIUM)

        for i in range(NUM_LOOPS):
            #req = uvm_create(req, self.m_sequencer)
            uvm_info("sequenceA", "Starting item " + str(i), UVM_MEDIUM)
            req = bus_req("my_req")
            await Timer(10, "NS")

            req.addr = (self.my_id * NUM_LOOPS) + i
            req.data = self.my_id + i + 55
            req.op   = BUS_WRITE

            if self.m_sequencer is None:
                raise Exception("Null sequencer CRASH")

            # uvm_send(req)
            await self.start_item(req)
            await self.finish_item(req)
            uvm_info("sequenceA", "getting response1 for " + str(i), UVM_MEDIUM)
            rsp = []
            await self.get_response(rsp)
            uvm_info("sequenceA", "got response1 for " + str(i) + ": " +
                    rsp[0].convert2string(), UVM_MEDIUM)

            #uvm_create(req)
            req = bus_req("my_req")
            req.addr = (self.my_id * NUM_LOOPS) + i
            req.data = 0
            req.op   = BUS_READ

            #uvm_send(req)
            await self.start_item(req)
            await self.finish_item(req)
            uvm_info("sequenceA", "getting response2 for " + str(i), UVM_MEDIUM)
            rsp = []
            await self.get_response(rsp)

            rsp = rsp[0]
            if rsp.data != (self.my_id + i + 55):
                uvm_error("SequenceA", sv.sformatf("Error, addr: %0d, expected data: %0d, actual data: %0d",
                    req.addr, req.data, rsp.data))
        uvm_info("sequenceA", "Finishing sequence", UVM_MEDIUM)

#uvm_object_param_utils(sequenceA)
uvm_object_utils(sequenceA)


class env(UVMEnv):
    #local uvm_sequencer #(bus_req, bus_rsp) sqr
    #local my_driver #(bus_req, bus_rsp) drv 

    def __init__(self, name, parent):
        UVMEnv.__init__(self, name, parent)
        self.drv = None
        self.sqr = None
        self.error = False

    def build_phase(self, phase):
        UVMEnv.build_phase(self, phase)
        self.sqr = UVMSequencer("sequence_controller", self)
        # create and connect driver
        self.drv = my_driver("my_driver", self)
        uvm_info("ENV", "Connecting ports now here", UVM_MEDIUM)
        self.drv.seq_item_port.connect(self.sqr.seq_item_export)

    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        fork_procs = []
        for i in range(NUM_SEQS):
            the_sequence = sequenceA("sequence_" + str(i))
            fork_procs.append(cocotb.fork(the_sequence.start(self.sqr, None)))
        join_list = list(map(lambda t: t.join(), fork_procs))
        await Combine(*join_list)
        phase.drop_objection(self)

    def check_phase(self, phase):
        if self.drv.nitems != (2 * NUM_SEQS * NUM_LOOPS):
            self.error = True
            self.uvm_report_error('env', sv.sformatf("nitems: Exp %0d, Got %0d",
                NUM_SEQS * NUM_LOOPS, self.drv.nitems))


uvm_component_utils(env)


@cocotb.test()
async def test_module_top(dut):
    UVMPhase.m_phase_trace = True
    e = env("env_name", parent=None)
    uvm_info("top","In top initial block", UVM_MEDIUM)
    await run_test()
    sim_time = sim_time = get_sim_time('ns')

    if sim_time == 0:
        raise Exception('Sim time has not progressed')
    else:
        if sim_time < 50.0:
            msg = "Exp: > 50, got: " + str(sim_time)
            raise Exception('Wrong sim time at the end: ' + msg)
    if e.error is True:
        raise Exception('Env had errors')




import unittest

from uvm.base.uvm_report_message import UVMReportMessage
from uvm.base.uvm_object_globals import UVM_LOG


class TestUVMReportMessage(unittest.TestCase):

    def test_message(self):
        msg = UVMReportMessage()
        cont = msg.get_element_container()
        self.assertEqual(cont.size(), 0)
        msg.add("msg", {1: 2}, UVM_LOG)
        self.assertEqual(cont.size(), 1)

    def test_msg_copy(self):
        msg = UVMReportMessage.new_report_message()
        msg.set_message("My message is this")
        msg.set_verbosity(111)
        msg.set_filename('log.txt')
        msg_copy = msg.clone()

        self.assertEqual(msg_copy.get_name(), msg.get_name())
        self.assertEqual(msg_copy.get_verbosity(), msg.get_verbosity())
        self.assertEqual(msg_copy.get_filename(), msg.get_filename())

        msg_copy = UVMReportMessage()
        msg_copy.copy(msg)
        self.assertEqual(msg_copy.get_message(), msg.get_message())


if __name__ == '__main__':
    unittest.main()



#`ifndef UVM_PHASE_DEFINES_SVH
#`define UVM_PHASE_DEFINES_SVH
#//
#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc. 
#//   Copyright 2011 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
#
#
#// uvm_root.svh uses these macros to simplify creation of all the phases.
#// they are only to be used for UVM builtin phases, because they are simple
#// delegate imps that call the corresponding methods on uvm_component.
#// Also, they declare classes (uvm_XXXXX_phase) and singleton instances (XXXXX_ph)
#
#// If you require more complex phase functors for your custom phase, code your
#// own imp class extending uvm_task/topdown/bottomup_phase base classes, following
#// the pattern of the macros below, but customize the exec_task() or exec_func()
#// contents to suit your enhanced functionality or derived component type/methods.
#// The uvm_user_xxx_phase() macros are provided for your convenience.
#
#
#`define m_uvm_task_phase(PHASE,COMP,PREFIX) \
#        class PREFIX``PHASE``_phase extends uvm_task_phase; \
#          virtual task exec_task(uvm_component comp, uvm_phase phase); \
#            COMP comp_; \
#            if ($cast(comp_,comp)) \
#              comp_.``PHASE``_phase(phase); \
#          endtask \
#          local static PREFIX``PHASE``_phase m_inst; \
#          static const string type_name = `"PREFIX``PHASE``_phase`"; \
#          static function PREFIX``PHASE``_phase get(); \
#            if(m_inst == null) begin \
#              m_inst = new; \
#            end \
#            return m_inst; \
#          endfunction \
#          protected function new(string name=`"PHASE`"); \
#            super.new(name); \
#          endfunction \
#          virtual function string get_type_name(); \
#            return type_name; \
#          endfunction \
#        endclass \
#        //PREFIX``PHASE``_phase PREFIX``PHASE``_ph = PREFIX``PHASE``_phase::get();
#
#`define m_uvm_topdown_phase(PHASE,COMP,PREFIX) \
#        class PREFIX``PHASE``_phase extends uvm_topdown_phase; \
#          virtual function void exec_func(uvm_component comp, uvm_phase phase); \
#            COMP comp_; \
#            if ($cast(comp_,comp)) \
#              comp_.``PHASE``_phase(phase); \
#          endfunction \
#          local static PREFIX``PHASE``_phase m_inst; \
#          static const string type_name = `"PREFIX``PHASE``_phase`"; \
#          static function PREFIX``PHASE``_phase get(); \
#            if(m_inst == null) begin \
#              m_inst = new(); \
#            end \
#            return m_inst; \
#          endfunction \
#          protected function new(string name=`"PHASE`"); \
#            super.new(name); \
#          endfunction \
#          virtual function string get_type_name(); \
#            return type_name; \
#          endfunction \
#        endclass \
#        //PREFIX``PHASE``_phase PREFIX``PHASE``_ph = PREFIX``PHASE``_phase::get();
#
#`define m_uvm_bottomup_phase(PHASE,COMP,PREFIX) \
#        class PREFIX``PHASE``_phase extends uvm_bottomup_phase; \
#          virtual function void exec_func(uvm_component comp, uvm_phase phase); \
#            COMP comp_; \
#            if ($cast(comp_,comp)) \
#              comp_.``PHASE``_phase(phase); \
#          endfunction \
#          static PREFIX``PHASE``_phase m_inst; \
#          static const string type_name = `"PREFIX``PHASE``_phase`"; \
#          static function PREFIX``PHASE``_phase get(); \
#            if(m_inst == null) begin \
#              m_inst = new(); \
#            end \
#            return m_inst; \
#          endfunction \
#          protected function new(string name=`"PHASE`"); \
#            super.new(name); \
#          endfunction \
#          virtual function string get_type_name(); \
#            return type_name; \
#          endfunction \
#        endclass \
#        //PREFIX``PHASE``_phase PREFIX``PHASE``_ph = PREFIX``PHASE``_phase::get();
#
#`define uvm_builtin_task_phase(PHASE) \
#        `m_uvm_task_phase(PHASE,uvm_component,uvm_)
#
#`define uvm_builtin_topdown_phase(PHASE) \
#        `m_uvm_topdown_phase(PHASE,uvm_component,uvm_)
#
#`define uvm_builtin_bottomup_phase(PHASE) \
#        `m_uvm_bottomup_phase(PHASE,uvm_component,uvm_)
#
#
#`define uvm_user_task_phase(PHASE,COMP,PREFIX) \
#        `m_uvm_task_phase(PHASE,COMP,PREFIX)
#
#`define uvm_user_topdown_phase(PHASE,COMP,PREFIX) \
#        `m_uvm_topdown_phase(PHASE,COMP,PREFIX)
#
#`define uvm_user_bottomup_phase(PHASE,COMP,PREFIX) \
#        `m_uvm_bottomup_phase(PHASE,COMP,PREFIX)
#
#`endif



#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.seq import UVMSequencer
from uvm.tlm1 import UVMBlockingPeekPort
from uvm.macros import uvm_component_utils

#//------------------------------------------------------------------------------
#//
#// CLASS: ubus_master_sequencer
#//
#//------------------------------------------------------------------------------

class ubus_master_sequencer(UVMSequencer):

    def __init__(self, name, parent):
        UVMSequencer.__init__(self, name, parent)
        self.addr_ph_port = UVMBlockingPeekPort("addr_ph_port", self)
    #endclass : ubus_master_sequencer
uvm_component_utils(ubus_master_sequencer)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2009 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//
"""
Title: Virtual Registers

A virtual register is a collection of fields,
overlaid on top of a memory, usually in an array.
The semantics and layout of virtual registers comes from
an agreement between the software and the hardware,
not any physical structures in the DUT.
"""


#typedef class uvm_mem_region
#typedef class uvm_mem_mam

#typedef class uvm_vreg_cbs

# import cocotb
from ..base import (UVMObject, sv, UVMMailbox)
from ..macros import (uvm_error, uvm_object_utils, UVM_REG_DATA_WIDTH,
    uvm_fatal)
from ..base.uvm_callback import *
from ..base.uvm_object_globals import (UVM_MEDIUM)


class UVMVReg(UVMObject):
    """
    Class: uvm_vreg

    Virtual register abstraction base class

    A virtual register represents a set of fields that are
    logically implemented in consecutive memory locations.

    All virtual register accesses eventually turn into memory accesses.

    A virtual register array may be implemented on top of
    any memory abstraction class and possibly dynamically
    resized and/or relocated.
    """

    #
    #   `uvm_register_cb(uvm_vreg, uvm_vreg_cbs)





    #   local semaphore atomic;   // Field RMW operations must be atomic

    #   //
    #   // Group: Initialization
    #   //

    #   //
    #   // FUNCTION: new
    #   // Create a new instance and type-specific configuration
    #   //
    #   // Creates an instance of a virtual register abstraction class
    #   // with the specified name.
    #   //
    #   // ~n_bits~ specifies the total number of bits in a virtual register.
    #   // Not all bits need to be mapped to a virtual field.
    #   // This value is usually a multiple of 8.
    #   //
    #   extern function new(string       name,
    #                       int unsigned n_bits)
    def __init__(self, name, n_bits):
        super().__init__(name)

        if n_bits == 0:
            uvm_error("RegModel", sv.sformatf("Virtual register '%s' cannot have 0 bits",
                self.get_full_name()))
            n_bits = 1

        if n_bits > UVM_REG_DATA_WIDTH:
            uvm_error("RegModel", sv.sformatf(
                "Virtual register '%s' cannot have more than %0d bits (%0d)",
                self.get_full_name(), UVM_REG_DATA_WIDTH, n_bits))
            n_bits = UVM_REG_DATA_WIDTH
        self.n_bits = n_bits

        self.locked = 0
        self.parent = None
        self.n_used_bits = 0
        self.locked = 0
        self.fields = []  # UVMVRegField[$]
        self.size = 0  # number of vregs
        self.offset = 0  # Start of vreg[0]
        self.mem = None  # UVMMem,  Where is it implemented?
        self.incr = 0  # From start to start of next
        self.is_static = 0
        self.region = None  # UVMMemRegion Not None if implemented via MAM
        self.fname = ""
        self.lineno = 0
        self.read_in_progress = 0
        self.write_in_progress = 0


    #   //
    #   // Function: configure
    #   // Instance-specific configuration
    #   //
    #   // Specify the ~parent~ block of this virtual register array.
    #   // If one of the other parameters are specified, the virtual register
    #   // is assumed to be dynamic and can be later (re-)implemented using
    #   // the <uvm_vreg::implement()> method.
    #   //
    #   // If ~mem~ is specified, then the virtual register array is assumed
    #   // to be statically implemented in the memory corresponding to the specified
    #   // memory abstraction class and ~size~, ~offset~ and ~incr~
    #   // must also be specified.
    #   // Static virtual register arrays cannot be re-implemented.
    #   //
    #   extern function void configure(uvm_reg_block     parent,
    #                                  uvm_mem       mem    = None,
    #                                  longint unsigned  size   = 0,
    #                                  uvm_reg_addr_t    offset = 0,
    #                                  int unsigned      incr   = 0)
    def configure(self, parent, mem=None, size=0, offset=0, incr=0):
        self.parent = parent

        self.n_used_bits = 0

        if mem is not None:
           self.implement(size, mem, offset, incr)  # cast to 'void' removed
           self.is_static = 1
        else:
           self.mem = None
           self.is_static = 0

        self.parent.add_vreg(self)
        # self.atomic = new(1)
        self.atomic  = UVMMailbox(1)
        self.atomic.try_put(1)
        #endfunction: configure


    #   //
    #   // FUNCTION: implement
    #   // Dynamically implement, resize or relocate a virtual register array
    #   //
    #   // Implement an array of virtual registers of the specified
    #   // ~size~, in the specified memory and ~offset~.
    #   // If an offset increment is specified, each
    #   // virtual register is implemented at the specified offset increment
    #   // from the previous one.
    #   // If an offset increment of 0 is specified,
    #   // virtual registers are packed as closely as possible
    #   // in the memory.
    #   //
    #   // If no memory is specified, the virtual register array is
    #   // in the same memory, at the same base offset using the same
    #   // offset increment as originally implemented.
    #   // Only the number of virtual registers in the virtual register array
    #   // is modified.
    #   //
    #   // The initial value of the newly-implemented or
    #   // relocated set of virtual registers is whatever values
    #   // are currently stored in the memory now implementing them.
    #   //
    #   // Returns TRUE if the memory
    #   // can implement the number of virtual registers
    #   // at the specified base offset and offset increment.
    #   // Returns FALSE otherwise.
    #   //
    #   // The memory region used to implement a virtual register array
    #   // is reserved in the memory allocation manager associated with
    #   // the memory to prevent it from being allocated for another purpose.
    #   //
    #   extern virtual function bit implement(longint unsigned  n,
    #                                         uvm_mem       mem    = None,
    #                                         uvm_reg_addr_t    offset = 0,
    #                                         int unsigned      incr   = 0)
    def implement(self, n, mem=None, offset=0, incr=0):
        #   uvm_mem_region region
        region = None

        if n < 1:
            uvm_error("RegModel", sv.sformatf(
                "Attempting to implement virtual reg '%s' with a subscript less than one doesn't make sense",
                self.get_full_name()))
            return 0


        if mem is None:
            uvm_error("RegModel", sv.sformatf(
                "Attempting to implement virtual register \"%s\" using a None UVMMem reference",
                self.get_full_name()))
            return 0

        if self.is_static:
            uvm_error("RegModel", sv.sformatf(
                "Virtual register \"%s\" is static and cannot be dynamically implemented",
                self.get_full_name()))
            return 0

        if mem.get_block() != self.parent:
            uvm_error("RegModel", sv.sformatf(
                "Attempting to implement virtual register '%s' on memory '%s' in a different block",
                self.get_full_name(),
                mem.get_full_name()))
            return 0

        min_incr = (self.get_n_bytes()-1) / mem.get_n_bytes() + 1
        if (incr == 0):
            incr = min_incr
        if min_incr > incr:
            uvm_error("RegModel", sv.sformatf(
                "Virtual register \"%s\" increment is too small (%0d): Each" +
                " virtual register requires at least %0d locations in memory '%s'.",
                self.get_full_name(), incr,
                min_incr, mem.get_full_name()))
            return 0


        # Is the memory big enough for ya?
        if offset + (n * incr) > mem.get_size():
            uvm_error("RegModel", sv.sformatf(
                "Given Offset for Virtual register \"%s[%0d]\" is too big for memory %s@'h%0h",
                self.get_full_name(), n, mem.get_full_name(), offset))
            return 0


        region = mem.mam.reserve_region(offset,n*incr*mem.get_n_bytes())

        if region is None:
            uvm_error("RegModel", sv.sformatf(
                "Could not allocate a memory region for virtual reg '%s'", self.get_full_name()))
            return 0

        if self.mem is not None:
            uvm_info("RegModel", sv.sformatf(
                "Virtual register \"%s\" is being moved re-implemented from %s@'h%0h to %s@'h%0h",
                self.get_full_name(),
                self.mem.get_full_name(),
                self.offset,
                mem.get_full_name(), offset), UVM_MEDIUM)
            self.release_region()

        self.region = region
        self.mem    = mem
        self.size   = n
        self.offset = offset
        self.incr   = incr
        self.mem.Xadd_vregX(self)
        
        return 1


    #   //
    #   // FUNCTION: allocate
    #   // Randomly implement, resize or relocate a virtual register array
    #   //
    #   // Implement a virtual register array of the specified
    #   // size in a randomly allocated region of the appropriate size
    #   // in the address space managed by the specified memory allocation manager.
    #   // If a memory allocation policy is specified, it is passed to the
    #   // uvm_mem_mam::request_region() method.
    #   //
    #   // The initial value of the newly-implemented
    #   // or relocated set of virtual registers is whatever values are
    #   // currently stored in the
    #   // memory region now implementing them.
    #   //
    #   // Returns a reference to a <uvm_mem_region> memory region descriptor
    #   // if the memory allocation manager was able to allocate a region
    #   // that can implement the virtual register array with the specified allocation policy.
    #   // Returns ~null~ otherwise.
    #   //
    #   // A region implementing a virtual register array
    #   // must not be released using the <uvm_mem_mam::release_region()> method.
    #   // It must be released using the <uvm_vreg::release_region()> method.
    #   //
    #   extern virtual function uvm_mem_region allocate(longint unsigned   n,
    #                                                   uvm_mem_mam        mam,
    #                                                   uvm_mem_mam_policy alloc = None)
    #
    #   //
    #   // FUNCTION: get_region
    #   // Get the region where the virtual register array is implemented
    #   //
    #   // Returns a reference to the <uvm_mem_region> memory region descriptor
    #   // that implements the virtual register array.
    #   //
    #   // Returns ~null~ if the virtual registers array
    #   // is not currently implemented.
    #   // A region implementing a virtual register array
    #   // must not be released using the <uvm_mem_mam::release_region()> method.
    #   // It must be released using the <uvm_vreg::release_region()> method.
    #   //
    #   extern def get_region(self):
    #
    #   //
    #   // FUNCTION: release_region
    #   // Dynamically un-implement a virtual register array
    #   //
    #   // Release the memory region used to implement a virtual register array
    #   // and return it to the pool of available memory
    #   // that can be allocated by the memory's default allocation manager.
    #   // The virtual register array is subsequently considered as unimplemented
    #   // and can no longer be accessed.
    #   //
    #   // Statically-implemented virtual registers cannot be released.
    #   //
    #   extern def release_region(self):
    #
    #
    #   /*local*/ extern virtual function void set_parent(uvm_reg_block parent);
    #   /*local*/ extern function void Xlock_modelX();
    #

    #   /*local*/ extern function void add_field(uvm_vreg_field field);
    def add_field(self, field):
        offset = 0
        idx = 0

        if self.locked:
            uvm_error("RegModel", "Cannot add virtual field to locked virtual register model")
            return

        if field is None:
            uvm_fatal("RegModel", "Attempting to register NULL virtual field")

        # Store fields in LSB to MSB order
        offset = field.get_lsb_pos_in_register()
        idx = -1
        for i in range(len(self.fields)):
            f = self.fields[i]
            if offset < f.get_lsb_pos_in_register():
                j = i
                self.fields.insert(j, field)
                idx = i
                break

        if idx < 0:
            self.fields.append(field)
            idx = len(self.fields) - 1

        self.n_used_bits += field.get_n_bits()

        # Check if there are too many fields in the register
        if self.n_used_bits > self.n_bits:
            uvm_error("RegModel", sv.sformatf(
                "Virtual fields use more bits (%0d) than available in virtual reg \"%s\" (%0d)",
                self.n_used_bits, self.get_full_name(), self.n_bits))

        # Check if there are overlapping fields
        if idx > 0:
            if (self.fields[idx-1].get_lsb_pos_in_register() +
                    self.fields[idx-1].get_n_bits() > offset):
                uvm_error("RegModel", sv.sformatf("Field %s overlaps field %s in virtual reg \"%s\"",
                    self.fields[idx-1].get_name(),
                    field.get_name(),
                    self.get_full_name()))

        if idx < len(self.fields) - 1:
            if (offset + field.get_n_bits() >
                    self.fields[idx+1].get_lsb_pos_in_register()):
                uvm_error("RegModel", sv.sformatf("Field %s overlaps field %s in virtual reg \"%s\"",
                    field.get_name(),
                    self.fields[idx+1].get_name(),
                    self.get_full_name()))


    #   /*local*/ extern task XatomicX(bit on);

    #
    #   //
    #   // Group: Introspection
    #   //
    #
    #   //
    #   // Function: get_name
    #   // Get the simple name
    #   //
    #   // Return the simple object name of this register.
    #   //
    #
    #   //
    #   // Function: get_full_name
    #   // Get the hierarchical name
    #   //
    #   // Return the hierarchal name of this register.
    #   // The base of the hierarchical name is the root block.
    #   //
    #   extern def string        get_full_name(self):
    def get_full_name(self):
        blk = None  # uvm_reg_block
        get_full_name = self.get_name()

        # Do not include top-level name in full name
        blk = self.get_block()
        if blk is None:
            return get_full_name
        if blk.get_parent() is None:
            return get_full_name

        get_full_name = self.parent.get_full_name() + "." + get_full_name
        return get_full_name


    #   //
    #   // FUNCTION: get_parent
    #   // Get the parent block
    #   //
    #   extern def get_parent(self):
    def get_parent(self):
        return self.parent


    #   extern def get_block(self):
    def get_block(self):
        return self.parent


    #   //
    #   // FUNCTION: get_memory
    #   // Get the memory where the virtual register array is implemented
    #   //
    #   extern def get_memory(self):
    #
    #   //
    #   // Function: get_n_maps
    #   // Returns the number of address maps this virtual register array is mapped in
    #   //
    #   extern def int             get_n_maps      (self):
    #
    #   //
    #   // Function: is_in_map
    #   // Return TRUE if this virtual register array is in the specified address ~map~
    #   //
    #   extern def         bit             is_in_map       (self,uvm_reg_map map):
    #
    #   //
    #   // Function: get_maps
    #   // Returns all of the address ~maps~ where this virtual register array is mapped
    #   //
    #   extern def void            get_maps        (self,ref uvm_reg_map maps[$]):
    #
    #   //
    #   // FUNCTION: get_rights
    #   // Returns the access rights of this virtual register array
    #   //
    #   // Returns "RW", "RO" or "WO".
    #   // The access rights of a virtual register array is always "RW",
    #   // unless it is implemented in a shared memory
    #   // with access restriction in a particular address map.
    #   //
    #   // If no address map is specified and the memory is mapped in only one
    #   // address map, that address map is used. If the memory is mapped
    #   // in more than one address map, the default address map of the
    #   // parent block is used.
    #   //
    #   // If an address map is specified and
    #   // the memory is not mapped in the specified
    #   // address map, an error message is issued
    #   // and "RW" is returned.
    #   //
    #   extern def get_rights(self,uvm_reg_map map = None):
    #
    #   //
    #   // FUNCTION: get_access
    #   // Returns the access policy of the virtual register array
    #   // when written and read via an address map.
    #   //
    #   // If the memory implementing the virtual register array
    #   // is mapped in more than one address map,
    #   // an address ~map~ must be specified.
    #   // If access restrictions are present when accessing a memory
    #   // through the specified address map, the access mode returned
    #   // takes the access restrictions into account.
    #   // For example, a read-write memory accessed
    #   // through an address map with read-only restrictions would return "RO".
    #   //
    #   extern def get_access(self,uvm_reg_map map = None):
    #

    #   //
    #   // FUNCTION: get_size
    #   // Returns the size of the virtual register array.
    #   //
    #   extern def int unsigned get_size(self):
    def get_size(self) -> int:
        if self.size == 0:
            uvm_error("RegModel", sv.sformatf(
                "Cannot call uvm_vreg::get_size() on unimplemented virtual register \"%s\"",
                self.get_full_name()))
            return 0
        return self.size


    #   //
    #   // FUNCTION: get_n_bytes
    #   // Returns the width, in bytes, of a virtual register.
    #   //
    #   // The width of a virtual register is always a multiple of the width
    #   // of the memory locations used to implement it.
    #   // For example, a virtual register containing two 1-byte fields
    #   // implemented in a memory with 4-bytes memory locations is 4-byte wide.
    #   //
    #   extern def int unsigned get_n_bytes(self):
    def get_n_bytes(self) -> int:
        return int((self.n_bits-1) / 8) + 1


    #   //
    #   // FUNCTION: get_n_memlocs
    #   // Returns the number of memory locations used
    #   // by a single virtual register.
    #   //
    #   extern def int unsigned get_n_memlocs(self):
    #
    #   //
    #   // FUNCTION: get_incr
    #   // Returns the number of memory locations
    #   // between two individual virtual registers in the same array.
    #   //
    #   extern def int unsigned get_incr(self):
    #
    #   //
    #   // FUNCTION: get_fields
    #   // Return the virtual fields in this virtual register
    #   //
    #   // Fills the specified array with the abstraction class
    #   // for all of the virtual fields contained in this virtual register.
    #   // Fields are ordered from least-significant position to most-significant
    #   // position within the register.
    #   //
    #   extern def get_fields(self,ref uvm_vreg_field fields[$]):
    #
    #   //
    #   // FUNCTION: get_field_by_name
    #   // Return the named virtual field in this virtual register
    #   //
    #   // Finds a virtual field with the specified name in this virtual register
    #   // and returns its abstraction class.
    #   // If no fields are found, returns ~null~.
    #   //
    #   extern def get_field_by_name(self,string name):
    #
    #   //
    #   // FUNCTION: get_offset_in_memory
    #   // Returns the offset of a virtual register
    #   //
    #   // Returns the base offset of the specified virtual register,
    #   // in the overall address space of the memory
    #   // that implements the virtual register array.
    #   //
    #   extern def uvm_reg_addr_t  get_offset_in_memory(self,longint unsigned idx):
    #
    #   //
    #   // FUNCTION: get_address
    #   // Returns the base external physical address of a virtual register
    #   //
    #   // Returns the base external physical address of the specified
    #   // virtual register if accessed through the specified address ~map~.
    #   //
    #   // If no address map is specified and the memory implementing
    #   // the virtual register array is mapped in only one
    #   // address map, that address map is used. If the memory is mapped
    #   // in more than one address map, the default address map of the
    #   // parent block is used.
    #   //
    #   // If an address map is specified and
    #   // the memory is not mapped in the specified
    #   // address map, an error message is issued.
    #   //
    #   extern virtual function uvm_reg_addr_t  get_address(longint unsigned idx,
    #                                                       uvm_reg_map map = None)
    #
    #   //
    #   // Group: HDL Access
    #   //
    #
    #   //
    #   // TASK: write
    #   // Write the specified value in a virtual register
    #   //
    #   // Write ~value~ in the DUT memory location(s) that implements
    #   // the virtual register array that corresponds to this
    #   // abstraction class instance using the specified access
    #   // ~path~.
    #   //
    #   // If the memory implementing the virtual register array
    #   // is mapped in more than one address map,
    #   // an address ~map~ must be
    #   // specified if a physical access is used (front-door access).
    #   //
    #   // The operation is eventually mapped into set of
    #   // memory-write operations at the location where the virtual register
    #   // specified by ~idx~ in the virtual register array is implemented.
    #   //
    #   extern def write(self,input  longint unsigned   idx,
    #                             output uvm_status_e  status,
    #                             input  uvm_reg_data_t     value,
    #                             input  uvm_path_e    path = UVM_DEFAULT_PATH,
    #                             input  uvm_reg_map     map = None,
    #                             input  uvm_sequence_base  parent = None,
    #                             input  uvm_object         extension = None,
    #                             input  string             fname = "",
    #                             input  int                lineno = 0)
    #
    #   //
    #   // TASK: read
    #   // Read the current value from a virtual register
    #   //
    #   // Read from the DUT memory location(s) that implements
    #   // the virtual register array that corresponds to this
    #   // abstraction class instance using the specified access
    #   // ~path~ and return the readback ~value~.
    #   //
    #   // If the memory implementing the virtual register array
    #   // is mapped in more than one address map,
    #   // an address ~map~ must be
    #   // specified if a physical access is used (front-door access).
    #   //
    #   // The operation is eventually mapped into set of
    #   // memory-read operations at the location where the virtual register
    #   // specified by ~idx~ in the virtual register array is implemented.
    #   //
    #   extern def read(self,input  longint unsigned    idx,
    #                            output uvm_status_e   status,
    #                            output uvm_reg_data_t      value,
    #                            input  uvm_path_e     path = UVM_DEFAULT_PATH,
    #                            input  uvm_reg_map      map = None,
    #                            input  uvm_sequence_base   parent = None,
    #                            input  uvm_object          extension = None,
    #                            input  string              fname = "",
    #                            input  int                 lineno = 0)
    #
    #   //
    #   // TASK: poke
    #   // Deposit the specified value in a virtual register
    #   //
    #   // Deposit ~value~ in the DUT memory location(s) that implements
    #   // the virtual register array that corresponds to this
    #   // abstraction class instance using the memory backdoor access.
    #   //
    #   // The operation is eventually mapped into set of
    #   // memory-poke operations at the location where the virtual register
    #   // specified by ~idx~ in the virtual register array is implemented.
    #   //
    #   extern def poke(self,input  longint unsigned    idx,
    #                            output uvm_status_e   status,
    #                            input  uvm_reg_data_t      value,
    #                            input  uvm_sequence_base   parent = None,
    #                            input  uvm_object          extension = None,
    #                            input  string              fname = "",
    #                            input  int                 lineno = 0)
    #
    #   //
    #   // TASK: peek
    #   // Sample the current value in a virtual register
    #   //
    #   // Sample the DUT memory location(s) that implements
    #   // the virtual register array that corresponds to this
    #   // abstraction class instance using the memory backdoor access,
    #   // and return the sampled ~value~.
    #   //
    #   // The operation is eventually mapped into set of
    #   // memory-peek operations at the location where the virtual register
    #   // specified by ~idx~ in the virtual register array is implemented.
    #   //
    #   extern def peek(self,input  longint unsigned    idx,
    #                            output uvm_status_e   status,
    #                            output uvm_reg_data_t      value,
    #                            input  uvm_sequence_base   parent = None,
    #                            input  uvm_object          extension = None,
    #                            input  string              fname = "",
    #                            input  int                 lineno = 0)
    #
    #   //
    #   // Function: reset
    #   // Reset the access semaphore
    #   //
    #   // Reset the semaphore that prevents concurrent access
    #   // to the virtual register.
    #   // This semaphore must be explicitly reset if a thread accessing
    #   // this virtual register array was killed in before the access
    #   // was completed
    #   //
    #   extern def reset(self,string kind = "HARD"):
    #
    #
    #   //
    #   // Group: Callbacks
    #   //
    #
    #   //
    #   // TASK: pre_write
    #   // Called before virtual register write.
    #   //
    #   // If the specified data value, access ~path~ or address ~map~ are modified,
    #   // the updated data value, access path or address map will be used
    #   // to perform the virtual register operation.
    #   //
    #   // The registered callback methods are invoked after the invocation
    #   // of this method.
    #   // All register callbacks are executed after the corresponding
    #   // field callbacks
    #   // The pre-write virtual register and field callbacks are executed
    #   // before the corresponding pre-write memory callbacks
    #   //
    #@cocotb.coroutine
    #   def pre_write(self,longint unsigned     idx,
    #                          ref uvm_reg_data_t   wdat,
    #                          ref uvm_path_e  path,
    #                          ref uvm_reg_map      map)
    #   endtask: pre_write
    #
    #   //
    #   // TASK: post_write
    #   // Called after virtual register write.
    #   //
    #   // If the specified ~status~ is modified,
    #   // the updated status will be
    #   // returned by the virtual register operation.
    #   //
    #   // The registered callback methods are invoked before the invocation
    #   // of this method.
    #   // All register callbacks are executed before the corresponding
    #   // field callbacks
    #   // The post-write virtual register and field callbacks are executed
    #   // after the corresponding post-write memory callbacks
    #   //
    #@cocotb.coroutine
    #   def post_write(self,longint unsigned       idx,
    #                           uvm_reg_data_t         wdat,
    #                           uvm_path_e        path,
    #                           uvm_reg_map            map,
    #                           ref uvm_status_e  status)
    #   endtask: post_write
    #
    #   //
    #   // TASK: pre_read
    #   // Called before virtual register read.
    #   //
    #   // If the specified access ~path~ or address ~map~ are modified,
    #   // the updated access path or address map will be used to perform
    #   // the register operation.
    #   //
    #   // The registered callback methods are invoked after the invocation
    #   // of this method.
    #   // All register callbacks are executed after the corresponding
    #   // field callbacks
    #   // The pre-read virtual register and field callbacks are executed
    #   // before the corresponding pre-read memory callbacks
    #   //
    #@cocotb.coroutine
    #   def pre_read(self,longint unsigned     idx,
    #                         ref uvm_path_e  path,
    #                         ref uvm_reg_map      map)
    #   endtask: pre_read
    #
    #   //
    #   // TASK: post_read
    #   // Called after virtual register read.
    #   //
    #   // If the specified readback data or ~status~ is modified,
    #   // the updated readback data or status will be
    #   // returned by the register operation.
    #   //
    #   // The registered callback methods are invoked before the invocation
    #   // of this method.
    #   // All register callbacks are executed before the corresponding
    #   // field callbacks
    #   // The post-read virtual register and field callbacks are executed
    #   // after the corresponding post-read memory callbacks
    #   //
    #@cocotb.coroutine
    #   def post_read(self,longint unsigned       idx,
    #                          ref uvm_reg_data_t     rdat,
    #                          input uvm_path_e  path,
    #                          input uvm_reg_map      map,
    #                          ref uvm_status_e  status)
    #   endtask: post_read
    #
    #   extern def do_print(self,uvm_printer printer):
    #   extern virtual function string convert2string
    #   extern def clone(self):
    #   extern def do_copy(self,uvm_object rhs):
    #   extern virtual function bit do_compare (uvm_object  rhs,
    #                                          uvm_comparer comparer)
    #   extern def do_pack(self,uvm_packer packer):
    #   extern def do_unpack(self,uvm_packer packer):
    #
    #endclass: uvm_vreg
#
#
#
#//------------------------------------------------------------------------------
#// Class: uvm_vreg_cbs
#//
#// Pre/post read/write callback facade class
#//
#//------------------------------------------------------------------------------
#
#class uvm_vreg_cbs(uvm_callback):
    #
    #   string fname
    #   int    lineno
    #
    #   def __init__(self, name = "uvm_reg_cbs")
    #      super().__init__(name)
    #   endfunction
    #
    #
    #   //
    #   // Task: pre_write
    #   // Callback called before a write operation.
    #   //
    #   // The registered callback methods are invoked after the invocation
    #   // of the <uvm_vreg::pre_write()> method.
    #   // All virtual register callbacks are executed after the corresponding
    #   // virtual field callbacks
    #   // The pre-write virtual register and field callbacks are executed
    #   // before the corresponding pre-write memory callbacks
    #   //
    #   // The written value ~wdat~, access ~path~ and address ~map~,
    #   // if modified, modifies the actual value, access path or address map
    #   // used in the virtual register operation.
    #   //
    #@cocotb.coroutine
    #   def pre_write(self,uvm_vreg         rg,
    #                          longint unsigned     idx,
    #                          ref uvm_reg_data_t   wdat,
    #                          ref uvm_path_e  path,
    #                          ref uvm_reg_map   map)
    #   endtask: pre_write
    #
    #
    #   //
    #   // TASK: post_write
    #   // Called after register write.
    #   //
    #   // The registered callback methods are invoked before the invocation
    #   // of the <uvm_reg::post_write()> method.
    #   // All register callbacks are executed before the corresponding
    #   // virtual field callbacks
    #   // The post-write virtual register and field callbacks are executed
    #   // after the corresponding post-write memory callbacks
    #   //
    #   // The ~status~ of the operation,
    #   // if modified, modifies the actual returned status.
    #   //
    #@cocotb.coroutine
    #   def post_write(self,uvm_vreg           rg,
    #                           longint unsigned       idx,
    #                           uvm_reg_data_t         wdat,
    #                           uvm_path_e        path,
    #                           uvm_reg_map         map,
    #                           ref uvm_status_e  status)
    #   endtask: post_write
    #
    #
    #   //
    #   // TASK: pre_read
    #   // Called before register read.
    #   //
    #   // The registered callback methods are invoked after the invocation
    #   // of the <uvm_reg::pre_read()> method.
    #   // All register callbacks are executed after the corresponding
    #   // virtual field callbacks
    #   // The pre-read virtual register and field callbacks are executed
    #   // before the corresponding pre-read memory callbacks
    #   //
    #   // The access ~path~ and address ~map~,
    #   // if modified, modifies the actual access path or address map
    #   // used in the register operation.
    #   //
    #@cocotb.coroutine
    #   def pre_read(self,uvm_vreg         rg,
    #                         longint unsigned     idx,
    #                         ref uvm_path_e  path,
    #                         ref uvm_reg_map   map)
    #   endtask: pre_read
    #
    #
    #   //
    #   // TASK: post_read
    #   // Called after register read.
    #   //
    #   // The registered callback methods are invoked before the invocation
    #   // of the <uvm_reg::post_read()> method.
    #   // All register callbacks are executed before the corresponding
    #   // virtual field callbacks
    #   // The post-read virtual register and field callbacks are executed
    #   // after the corresponding post-read memory callbacks
    #   //
    #   // The readback value ~rdat~ and the ~status~ of the operation,
    #   // if modified, modifies the actual returned readback value and status.
    #   //
    #@cocotb.coroutine
    #   def post_read(self,uvm_vreg           rg,
    #                          longint unsigned       idx,
    #                          ref uvm_reg_data_t     rdat,
    #                          input uvm_path_e  path,
    #                          input uvm_reg_map   map,
    #                          ref uvm_status_e  status)
    #   endtask: post_read
    #endclass: uvm_vreg_cbs
#
#
#//
#// Type: uvm_vreg_cb
#// Convenience callback type declaration
#//
#// Use this declaration to register virtual register callbacks rather than
#// the more verbose parameterized class
#//
#typedef uvm_callbacks#(uvm_vreg, uvm_vreg_cbs) uvm_vreg_cb
#
#//
#// Type: uvm_vreg_cb_iter
#// Convenience callback iterator type declaration
#//
#// Use this declaration to iterate over registered virtual register callbacks
#// rather than the more verbose parameterized class
#//
#typedef uvm_callback_iter#(uvm_vreg, uvm_vreg_cbs) uvm_vreg_cb_iter
#
#
#
#//------------------------------------------------------------------------------
#// IMPLEMENTATION
#//------------------------------------------------------------------------------
#
#
#
#
#
#def void uvm_vreg::Xlock_modelX(self):
#   if (self.locked) return
#
#   self.locked = 1
#endfunction: Xlock_modelX
#
#
#
#
#@cocotb.coroutine
#def uvm_vreg::XatomicX(self,bit on):  # task
#   if (on) self.atomic.get(1)
#   else begin
#      // Maybe a key was put back in by a spurious call to reset()
#      self.atomic.try_get(1)  # cast to 'void' removed
#      self.atomic.put(1)
#   end
#endtask: XatomicX
#
#
#def void uvm_vreg::reset(self,string kind = "HARD"):
#   // Put back a key in the semaphore if it is checked out
#   // in case a thread was killed during an operation
#   self.atomic.try_get(1)  # cast to 'void' removed
#   self.atomic.put(1)
#endfunction: reset
#
#
#
#def void uvm_vreg::set_parent(self,uvm_reg_block parent):
#   self.parent = parent
#endfunction: set_parent
#
#
#
#function uvm_mem_region uvm_vreg::allocate(longint unsigned   n,
#                                           uvm_mem_mam        mam,
#                                           uvm_mem_mam_policy alloc=None)
#
#   uvm_mem mem
#
#   if(n < 1)
#   begin
#     uvm_error("RegModel", sv.sformatf("Attempting to implement virtual register \"%s\" with a subscript less than one doesn't make sense",self.get_full_name()))
#      return None
#   end
#
#   if (mam is None):
#      uvm_error("RegModel", sv.sformatf("Attempting to implement virtual register \"%s\" using a NULL uvm_mem_mam reference", self.get_full_name()))
#      return None
#   end
#
#   if (self.is_static):
#      uvm_error("RegModel", sv.sformatf("Virtual register \"%s\" is static and cannot be dynamically allocated", self.get_full_name()))
#      return None
#   end
#
#   mem = mam.get_memory()
#   if (mem.get_block() != self.parent):
#      uvm_error("RegModel", sv.sformatf("Attempting to allocate virtual register \"%s\" on memory \"%s\" in a different block",
#                                     self.get_full_name(),
#                                     mem.get_full_name()))
#      return None
#   end
#
#   begin
#      int min_incr = (self.get_n_bytes()-1) / mem.get_n_bytes() + 1
#      if (incr == 0) incr = min_incr
#      if (min_incr < incr):
#         uvm_error("RegModel", sv.sformatf("Virtual register \"%s\" increment is too small (%0d): Each virtual register requires at least %0d locations in memory \"%s\".",
#                                        self.get_full_name(), incr,
#                                        min_incr, mem.get_full_name()))
#         return None
#      end
#   end
#
#   // Need memory at least of size num_vregs*sizeof(vreg) in bytes.
#   allocate = mam.request_region(n*incr*mem.get_n_bytes(), alloc)
#   if (allocate is None):
#      uvm_error("RegModel", sv.sformatf("Could not allocate a memory region for virtual register \"%s\"", self.get_full_name()))
#      return None
#   end
#
#   if (self.mem is not None):
#     uvm_info("RegModel", sv.sformatf("Virtual register \"%s\" is being moved from %s@'h%0h to %s@'h%0h",
#                                self.get_full_name(),
#                                self.mem.get_full_name(),
#                                self.offset,
#                                mem.get_full_name(),
#                                allocate.get_start_offset()),UVM_MEDIUM)
#
#      self.release_region()
#   end
#
#   self.region = allocate
#
#   self.mem    = mam.get_memory()
#   self.offset = allocate.get_start_offset()
#   self.size   = n
#   self.incr   = incr
#
#   self.mem.Xadd_vregX(self)
#endfunction: allocate
#
#
#def uvm_mem_region uvm_vreg::get_region(self):
#   return self.region
#endfunction: get_region
#
#
#def void uvm_vreg::release_region(self):
#   if (self.is_static):
#      uvm_error("RegModel", sv.sformatf("Virtual register \"%s\" is static and cannot be dynamically released", self.get_full_name()))
#      return
#   end
#
#   if (self.mem is not None)
#      self.mem.Xdelete_vregX(self)
#
#   if (self.region is not None):
#      self.region.release_region()
#   end
#
#   self.region = None
#   self.mem    = None
#   self.size   = 0
#   self.offset = 0
#
#   self.reset()
#endfunction: release_region
#
#
#def uvm_mem uvm_vreg::get_memory(self):
#   return self.mem
#endfunction: get_memory
#
#
#def uvm_reg_addr_t  uvm_vreg::get_offset_in_memory(self,longint unsigned idx):
#   if (self.mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot call uvm_vreg::get_offset_in_memory() on unimplemented virtual register \"%s\"",
#                                     self.get_full_name()))
#      return 0
#   end
#
#   return self.offset + idx * self.incr
#endfunction
#
#
#function uvm_reg_addr_t  uvm_vreg::get_address(longint unsigned idx,
#                                                   uvm_reg_map map = None)
#   if (self.mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot get address of of unimplemented virtual register \"%s\".", self.get_full_name()))
#      return 0
#   end
#
#   return self.mem.get_address(self.get_offset_in_memory(idx), map)
#endfunction: get_address
#
#
#
#
#
#def int unsigned uvm_vreg::get_n_memlocs(self):
#   if (self.mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot call uvm_vreg::get_n_memlocs() on unimplemented virtual register \"%s\"",
#                                     self.get_full_name()))
#      return 0
#   end
#
#   return (self.get_n_bytes()-1) / self.mem.get_n_bytes() + 1
#endfunction: get_n_memlocs
#
#
#def int unsigned uvm_vreg::get_incr(self):
#   if (self.incr == 0):
#      uvm_error("RegModel", sv.sformatf("Cannot call uvm_vreg::get_incr() on unimplemented virtual register \"%s\"",
#                                     self.get_full_name()))
#      return 0
#   end
#
#   return self.incr
#endfunction: get_incr
#
#
#def int uvm_vreg::get_n_maps(self):
#   if (self.mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot call uvm_vreg::get_n_maps() on unimplemented virtual register \"%s\"",
#                                     self.get_full_name()))
#      return 0
#   end
#
#   return self.mem.get_n_maps()
#endfunction: get_n_maps
#
#
#def void uvm_vreg::get_maps(self,ref uvm_reg_map maps[$]):
#   if (self.mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot call uvm_vreg::get_maps() on unimplemented virtual register \"%s\"",
#                                     self.get_full_name()))
#      return
#   end
#
#   self.mem.get_maps(maps)
#endfunction: get_maps
#
#
#def bit uvm_vreg::is_in_map(self,uvm_reg_map map):
#   if (self.mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot call uvm_vreg::is_in_map() on unimplemented virtual register \"%s\"",
#                                  self.get_full_name()))
#      return 0
#   end
#
#   return self.mem.is_in_map(map)
#endfunction
#
#
#def string uvm_vreg::get_access(self,uvm_reg_map map = None):
#   if (self.mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot call uvm_vreg::get_rights() on unimplemented virtual register \"%s\"",
#                                     self.get_full_name()))
#      return "RW"
#   end
#
#   return self.mem.get_access(map)
#endfunction: get_access
#
#
#def string uvm_vreg::get_rights(self,uvm_reg_map map = None):
#   if (self.mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot call uvm_vreg::get_rights() on unimplemented virtual register \"%s\"",
#                                     self.get_full_name()))
#      return "RW"
#   end
#
#   return self.mem.get_rights(map)
#endfunction: get_rights
#
#
#def void uvm_vreg::get_fields(self,ref uvm_vreg_field fields[$]):
#   foreach(self.fields[i])
#      fields.push_back(self.fields[i])
#endfunction: get_fields
#
#
#def uvm_vreg_field uvm_vreg::get_field_by_name(self,string name):
#   foreach (self.fields[i]):
#      if (self.fields[i].get_name() == name):
#         return self.fields[i]
#      end
#   end
#   uvm_warning("RegModel", sv.sformatf("Unable to locate field \"%s\" in virtual register \"%s\".",
#                                    name, self.get_full_name()))
#   get_field_by_name = None
#endfunction: get_field_by_name
#
#
#@cocotb.coroutine
#task uvm_vreg::write(input  longint unsigned   idx,
#                         output uvm_status_e  status,
#                         input  uvm_reg_data_t     value,
#                         input  uvm_path_e    path = UVM_DEFAULT_PATH,
#                         input  uvm_reg_map     map = None,
#                         input  uvm_sequence_base  parent = None,
#                         input  uvm_object         extension = None,
#                         input  string             fname = "",
#                         input  int                lineno = 0)
#   uvm_vreg_cb_iter cbs = new(self)
#
#   uvm_reg_addr_t  addr
#   uvm_reg_data_t  tmp
#   uvm_reg_data_t  msk
#   int lsb
#
#   self.write_in_progress = 1'b1
#   self.fname = fname
#   self.lineno = lineno
#   if (self.mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot write to unimplemented virtual register \"%s\".", self.get_full_name()))
#      status = UVM_NOT_OK
#      return
#   end
#
#   if (path == UVM_DEFAULT_PATH)
#     path = self.parent.get_default_path()
#
#   foreach (fields[i]):
#      uvm_vreg_field_cb_iter cbs = new(fields[i])
#      uvm_vreg_field f = fields[i]
#
#      lsb = f.get_lsb_pos_in_register()
#      msk = ((1<<f.get_n_bits())-1) << lsb
#      tmp = (value & msk) >> lsb
#
#      f.pre_write(idx, tmp, path, map)
#      for (uvm_vreg_field_cbs cb = cbs.first(); cb is not None
#           cb = cbs.next()):
#         cb.fname = self.fname
#         cb.lineno = self.lineno
#         cb.pre_write(f, idx, tmp, path, map)
#      end
#
#      value = (value & ~msk) | (tmp << lsb)
#   end
#   self.pre_write(idx, value, path, map)
#   for (uvm_vreg_cbs cb = cbs.first(); cb is not None
#        cb = cbs.next()):
#      cb.fname = self.fname
#      cb.lineno = self.lineno
#      cb.pre_write(self, idx, value, path, map)
#   end
#
#   addr = self.offset + (idx * self.incr)
#
#   lsb = 0
#   status = UVM_IS_OK
#   for (int i = 0; i < self.get_n_memlocs(); i++):
#      uvm_status_e s
#
#      msk = ((1<<(self.mem.get_n_bytes()*8))-1) << lsb
#      tmp = (value & msk) >> lsb
#      self.mem.write(s, addr + i, tmp, path, map , parent, , extension, fname, lineno)
#      if (s != UVM_IS_OK  and  s != UVM_HAS_X) status = s
#      lsb += self.mem.get_n_bytes() * 8
#   end
#
#   for (uvm_vreg_cbs cb = cbs.first(); cb is not None
#        cb = cbs.next()):
#      cb.fname = self.fname
#      cb.lineno = self.lineno
#      cb.post_write(self, idx, value, path, map, status)
#   end
#   self.post_write(idx, value, path, map, status)
#   foreach (fields[i]):
#      uvm_vreg_field_cb_iter cbs = new(fields[i])
#      uvm_vreg_field f = fields[i]
#
#      lsb = f.get_lsb_pos_in_register()
#      msk = ((1<<f.get_n_bits())-1) << lsb
#      tmp = (value & msk) >> lsb
#
#      for (uvm_vreg_field_cbs cb = cbs.first(); cb is not None
#           cb = cbs.next()):
#         cb.fname = self.fname
#         cb.lineno = self.lineno
#         cb.post_write(f, idx, tmp, path, map, status)
#      end
#      f.post_write(idx, tmp, path, map, status)
#
#      value = (value & ~msk) | (tmp << lsb)
#   end
#
#   uvm_info("RegModel", sv.sformatf("Wrote virtual register \"%s\"[%0d] via %s with: 'h%h",
#                              self.get_full_name(), idx,
#                              (path == UVM_FRONTDOOR) ? "frontdoor" : "backdoor",
#                              value),UVM_MEDIUM)
#
#   self.write_in_progress = 1'b0
#   self.fname = ""
#   self.lineno = 0
#
#endtask: write
#
#
#@cocotb.coroutine
#task uvm_vreg::read(input  longint unsigned   idx,
#                        output uvm_status_e  status,
#                        output uvm_reg_data_t     value,
#                        input  uvm_path_e    path = UVM_DEFAULT_PATH,
#                        input  uvm_reg_map     map = None,
#                        input  uvm_sequence_base  parent = None,
#                        input  uvm_object         extension = None,
#                        input  string             fname = "",
#                        input  int                lineno = 0)
#   uvm_vreg_cb_iter cbs = new(self)
#
#   uvm_reg_addr_t  addr
#   uvm_reg_data_t  tmp
#   uvm_reg_data_t  msk
#   int lsb
#   self.read_in_progress = 1'b1
#   self.fname = fname
#   self.lineno = lineno
#
#   if (self.mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot read from unimplemented virtual register \"%s\".", self.get_full_name()))
#      status = UVM_NOT_OK
#      return
#   end
#
#   if (path == UVM_DEFAULT_PATH)
#     path = self.parent.get_default_path()
#
#   foreach (fields[i]):
#      uvm_vreg_field_cb_iter cbs = new(fields[i])
#      uvm_vreg_field f = fields[i]
#
#      f.pre_read(idx, path, map)
#      for (uvm_vreg_field_cbs cb = cbs.first(); cb is not None
#           cb = cbs.next()):
#         cb.fname = self.fname
#         cb.lineno = self.lineno
#         cb.pre_read(f, idx, path, map)
#      end
#   end
#   self.pre_read(idx, path, map)
#   for (uvm_vreg_cbs cb = cbs.first(); cb is not None
#        cb = cbs.next()):
#      cb.fname = self.fname
#      cb.lineno = self.lineno
#      cb.pre_read(self, idx, path, map)
#   end
#
#   addr = self.offset + (idx * self.incr)
#
#   lsb = 0
#   value = 0
#   status = UVM_IS_OK
#   for (int i = 0; i < self.get_n_memlocs(); i++):
#      uvm_status_e s
#
#      self.mem.read(s, addr + i, tmp, path, map, parent, , extension, fname, lineno)
#      if (s != UVM_IS_OK  and  s != UVM_HAS_X) status = s
#
#      value |= tmp << lsb
#      lsb += self.mem.get_n_bytes() * 8
#   end
#
#   for (uvm_vreg_cbs cb = cbs.first(); cb is not None
#        cb = cbs.next()):
#      cb.fname = self.fname
#      cb.lineno = self.lineno
#      cb.post_read(self, idx, value, path, map, status)
#   end
#   self.post_read(idx, value, path, map, status)
#   foreach (fields[i]):
#      uvm_vreg_field_cb_iter cbs = new(fields[i])
#      uvm_vreg_field f = fields[i]
#
#      lsb = f.get_lsb_pos_in_register()
#
#      msk = ((1<<f.get_n_bits())-1) << lsb
#      tmp = (value & msk) >> lsb
#
#      for (uvm_vreg_field_cbs cb = cbs.first(); cb is not None
#           cb = cbs.next()):
#         cb.fname = self.fname
#         cb.lineno = self.lineno
#         cb.post_read(f, idx, tmp, path, map, status)
#      end
#      f.post_read(idx, tmp, path, map, status)
#
#      value = (value & ~msk) | (tmp << lsb)
#   end
#
#   uvm_info("RegModel", sv.sformatf("Read virtual register \"%s\"[%0d] via %s: 'h%h",
#                              self.get_full_name(), idx,
#                              (path == UVM_FRONTDOOR) ? "frontdoor" : "backdoor",
#                              value),UVM_MEDIUM)
#
#   self.read_in_progress = 1'b0
#   self.fname = ""
#   self.lineno = 0
#endtask: read
#
#
#@cocotb.coroutine
#task uvm_vreg::poke(input longint unsigned   idx,
#                        output uvm_status_e status,
#                        input  uvm_reg_data_t    value,
#                        input  uvm_sequence_base parent = None,
#                        input  uvm_object        extension = None,
#                        input  string            fname = "",
#                        input  int               lineno = 0)
#   uvm_reg_addr_t  addr
#   uvm_reg_data_t  tmp
#   uvm_reg_data_t  msk
#   int lsb
#   self.fname = fname
#   self.lineno = lineno
#
#   if (self.mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot poke in unimplemented virtual register \"%s\".", self.get_full_name()))
#      status = UVM_NOT_OK
#      return
#   end
#
#   addr = self.offset + (idx * self.incr)
#
#   lsb = 0
#   status = UVM_IS_OK
#   for (int i = 0; i < self.get_n_memlocs(); i++):
#      uvm_status_e s
#
#      msk = ((1<<(self.mem.get_n_bytes() * 8))-1) << lsb
#      tmp = (value & msk) >> lsb
#
#      self.mem.poke(status, addr + i, tmp, "", parent, extension, fname, lineno)
#      if (s != UVM_IS_OK  and  s != UVM_HAS_X) status = s
#
#      lsb += self.mem.get_n_bytes() * 8
#   end
#
#   uvm_info("RegModel", sv.sformatf("Poked virtual register \"%s\"[%0d] with: 'h%h",
#                              self.get_full_name(), idx, value),UVM_MEDIUM)
#   self.fname = ""
#   self.lineno = 0
#
#endtask: poke
#
#
#@cocotb.coroutine
#task uvm_vreg::peek(input longint unsigned   idx,
#                        output uvm_status_e status,
#                        output uvm_reg_data_t    value,
#                        input  uvm_sequence_base parent = None,
#                        input  uvm_object        extension = None,
#                        input  string            fname = "",
#                        input  int               lineno = 0)
#   uvm_reg_addr_t  addr
#   uvm_reg_data_t  tmp
#   uvm_reg_data_t  msk
#   int lsb
#   self.fname = fname
#   self.lineno = lineno
#
#   if (self.mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot peek in from unimplemented virtual register \"%s\".", self.get_full_name()))
#      status = UVM_NOT_OK
#      return
#   end
#
#   addr = self.offset + (idx * self.incr)
#
#   lsb = 0
#   value = 0
#   status = UVM_IS_OK
#   for (int i = 0; i < self.get_n_memlocs(); i++):
#      uvm_status_e s
#
#      self.mem.peek(status, addr + i, tmp, "", parent, extension, fname, lineno)
#      if (s != UVM_IS_OK  and  s != UVM_HAS_X) status = s
#
#      value |= tmp << lsb
#      lsb += self.mem.get_n_bytes() * 8
#   end
#
#   uvm_info("RegModel", sv.sformatf("Peeked virtual register \"%s\"[%0d]: 'h%h",
#                              self.get_full_name(), idx, value),UVM_MEDIUM)
#
#   self.fname = ""
#   self.lineno = 0
#
#endtask: peek
#
#
#def void uvm_vreg::do_print (self,uvm_printer printer):
#  super().do_print(printer)
#  printer.print_generic("initiator", parent.get_type_name(), -1, convert2string())
#endfunction
#
#def string uvm_vreg::convert2string(self):
#   string res_str
#   string t_str
#   bit with_debug_info
#   $sformat(convert2string, "Virtual register %s -- ",
#            self.get_full_name())
#
#   if (self.size == 0)
#     $sformat(convert2string, "%sunimplemented", convert2string)
#   else begin
#      uvm_reg_map maps[$]
#      mem.get_maps(maps)
#
#      $sformat(convert2string, "%s[%0d] in %0s['h%0h+'h%0h]\n", convert2string,
#             self.size, self.mem.get_full_name(), self.offset, self.incr);
#      foreach (maps[i]):
#        uvm_reg_addr_t  addr0 = self.get_address(0, maps[i])
#
#        $sformat(convert2string, "  Address in map '%s' -- @'h%0h+%0h",
#        maps[i].get_full_name(), addr0, self.get_address(1, maps[i]) - addr0)
#      end
#   end
#   foreach(self.fields[i]):
#      $sformat(convert2string, "%s\n%s", convert2string,
#               self.fields[i].convert2string())
#   end
#
#endfunction: convert2string
#
#
#
#//TODO - add fatal messages
#def uvm_object uvm_vreg::clone(self):
#  return None
#endfunction
#
#def void uvm_vreg::do_copy   (self,uvm_object rhs):
#endfunction
#
#function bit uvm_vreg::do_compare (uvm_object  rhs,
#                                        uvm_comparer comparer)
#  return 0
#endfunction
#
#def void uvm_vreg::do_pack (self,uvm_packer packer):
#endfunction
#
#def void uvm_vreg::do_unpack (self,uvm_packer packer):
#endfunction




#// -------------------------------------------------------------
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//


import cocotb
from cocotb.triggers import Timer
from uvm.macros import *
from uvm.base import UVMConfigDb
from uvm.reg import UVMRegSequence
from uvm.reg.uvm_reg_backdoor import UVMRegBackdoor
from uvm.reg.uvm_reg_model import (UVM_IS_OK, UVM_BACKDOOR)


class mem_backdoor(UVMRegBackdoor):

    def __init__(self, name="mem_backdoor"):
        super().__init__(name)
        self.dut = None
        self.mem_name = "DMA"

    
    async def write(self, rw):  # task
        await self.do_pre_write(rw)  # Required, as stated in uvm_reg_backdoor
        #uvm_fatal("RegModel", "UVMRegBackdoor::write() method has not been overloaded")
        print("rw.offset is " + str(rw.offset))
        self.dut.DMA[rw.offset] <= rw.value[0]
        rw.status = UVM_IS_OK
        await self.do_post_write(rw)  # Required, as stated in uvm_reg_backdoor


    def read_func(self, rw):
        # uvm_fatal("RegModel", "UVMRegBackdoor::read_func() method has not been overloaded")
        data = 0
        rw.status = UVM_IS_OK
        offset = rw.offset
        data = self.dut.DMA[offset]
        rw.value[0] = data
        #endfunction



class mem_test_seq(UVMRegSequence):

    def __init__(self, name="user_test_seq"):
        super().__init__(name)
        self.addr = None  # type: int
        self.data = None  # type: int
        self.dut = None

    
    async def body(self):
        arr = []
        if UVMConfigDb.get(None, "", "dut", arr):
            self.dut = arr[0]
        else:
            uvm_fatal("NO_DUT", "Could not find DUT from config_db")
        model = self.model
        dma_ram = model.DMA_RAM
        bkd = mem_backdoor("backdoor")
        dma_ram.set_backdoor(bkd)
        bkd.dut = self.dut

        for i in range(10):

            status = []
            data = 0x1234
            offset = i * 4
            await dma_ram.write(status, offset, data)
            await Timer(100, "NS")

            status = []
            data = []
            offset = i * 4
            await dma_ram.read(status, offset, data)
            exp = data[0]
            await Timer(100, "NS")
            status = []
            data = []
            await dma_ram.read(status, offset, data, path=UVM_BACKDOOR)
            if exp != data[0]:
                uvm_error("MEM_READ_ERR", "Exp: {}, Got: {}".format(exp, data[0]))

            # Test backdoor access with write/read operation
            ref_data = 0xBEEFFACE
            status = []
            await Timer(100, "NS")
            await dma_ram.write(status, offset, ref_data, path=UVM_BACKDOOR)
            status = []
            data = []
            await Timer(100, "NS")
            await dma_ram.read(status, offset, data, path=UVM_BACKDOOR)
            if data[0] != ref_data:
                uvm_fatal("MEM_DATA_ERR", "Exp: {}, Got: {}".format(ref_data, data[0]))
        await Timer(100, "NS")


uvm_object_utils(mem_test_seq)




import unittest

from uvm.tlm1 import UVMSeqItemPullExport


class TestUVMSeqItemPullExport(unittest.TestCase):

    def test_new(self):
        pull_exp = UVMSeqItemPullExport('pull_export', None)
        self.assertEqual(pull_exp.get_name(), 'pull_export')


if __name__ == '__main__':
    unittest.main()



#----------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http:#www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#----------------------------------------------------------------------
"""
Title: UVM Common Phases

The common phases are the set of function and task phases that all
`UVMComponent`s execute together.
All `UVMComponent`s are always synchronized
with respect to the common phases.

The names of the UVM phases (which will be returned by get_name() for a
phase instance) match the class names specified below with the "uvm_"
and "_phase" removed.  For example, the build phase corresponds to the
uvm_build_phase class below and has the name "build", which means that
the following can be used to call foo() at the end of the build phase
(after all lower levels have finished build)::

  def phase_ended(self, phase):
      if (phase.get_name()=="build"):
          self.foo()

The common phases are executed in the sequence they are specified below.
"""

import cocotb

from .uvm_topdown_phase import UVMTopdownPhase
from .uvm_bottomup_phase import UVMBottomupPhase
from .uvm_debug import uvm_debug
from .uvm_task_phase import UVMTaskPhase
#from .uvm_common_phases import *



class UVMBuildPhase(UVMTopdownPhase):
    """
    Class: UVMBuildPhase

    Create and configure of testbench structure
    `UVMTopdownPhase` that calls the `UVMComponent.build_phase` method.

    Upon entry:
      The top-level components have been instantiated under `UVMRoot`.
      Current simulation time is still equal to 0 but some "delta cycles" may have occurred

    Typical Uses:
      Instantiate sub-components.
      Instantiate register model.
      Get configuration values for the component being built.
      Set configuration values for sub-components.

    Exit Criteria:
      All `UVMComponent`s have been instantiated.
    """


    def __init__(self, name="build"):
        UVMTopdownPhase.__init__(self, name)

    def exec_func(self, comp, phase):
        comp.build_phase(phase)

    #local static uvm_build_phase m_inst
    m_inst = None
    type_name = "uvm_build_phase"

    @classmethod
    def get(cls):
        """
        Function: get
        Returns the singleton phase handle

        Returns:
            UVMBuildPhase:
        """
        uvm_debug(cls, 'get', 'called with ' + str(cls))
        if UVMBuildPhase.m_inst is None:
            uvm_debug(cls, 'get', "UVMBuildPhase is none. Returning new class")
            UVMBuildPhase.m_inst = UVMBuildPhase()
        return UVMBuildPhase.m_inst

    def get_type_name(self):
        return UVMBuildPhase.type_name


class UVMConnectPhase(UVMBottomupPhase):
    """
    Establish cross-component connections.

    `UVMBottomupPhase` that calls the
    `UVMComponent.connect_phase` method.

    Upon Entry:
    - All components have been instantiated.
    - Current simulation time is still equal to 0
      but some "delta cycles" may have occurred.

    Typical Uses:
    - Connect TLM ports and exports.
    - Connect TLM initiator sockets and target sockets.
    - Connect register model to adapter components.
    - Setup explicit phase domains.

    Exit Criteria:
    - All cross-component connections have been established.
    - All independent phase domains are set.

    """
    def exec_func(self, comp, phase):
        comp.connect_phase(phase)

    m_inst = None
    type_name = "uvm_connect_phase"

    @classmethod
    def get(cls):
        """
        Function: get
        Returns the singleton phase handle
        Returns:
        """
        if UVMConnectPhase.m_inst is None:
            UVMConnectPhase.m_inst = UVMConnectPhase()
        return UVMConnectPhase.m_inst

    def __init__(self, name="connect"):
        UVMBottomupPhase.__init__(self, name)

    def get_type_name(self):
        return UVMConnectPhase.type_name


class UVMEndOfElaborationPhase(UVMBottomupPhase):
    """
    Fine-tune the testbench.

    `UVMBottomupPhase` that calls the
    `UVMComponent.end_of_elaboration_phase` method.

    Upon Entry:
    - The verification environment has been completely assembled.
    - Current simulation time is still equal to 0
      but some "delta cycles" may have occurred.

    Typical Uses:
    - Display environment topology.
    - Open files.
    - Define additional configuration settings for components.

    Exit Criteria:
    - None.
    """
    def exec_func(self, comp, phase):
        comp.end_of_elaboration_phase(phase)

    m_inst = None
    type_name = "uvm_end_of_elaboration_phase"

    @classmethod
    def get(cls):
        """
        Function: get
        Returns the singleton phase handle
        Returns:
        """
        if UVMEndOfElaborationPhase.m_inst is None:
            UVMEndOfElaborationPhase.m_inst = UVMEndOfElaborationPhase()
        return UVMEndOfElaborationPhase.m_inst

    def __init__(self, name="end_of_elaboration"):
        UVMBottomupPhase.__init__(self, name)

    def get_type_name(self):
        return UVMEndOfElaborationPhase.type_name


class UVMStartofSimulationPhase(UVMBottomupPhase):
    """
    Get ready for DUT to be simulated.

    `UVMBottomupPhase` that calls the
    `UVMComponent.start_of_simulation_phase` method.

    Upon Entry:
      - Other simulation engines, debuggers, hardware assisted platforms and
        all other run-time tools have been started and synchronized.
      - The verification environment has been completely configured
        and is ready to start.
      - Current simulation time is still equal to 0
        but some "delta cycles" may have occurred.

    Typical Uses:
      - Display environment topology
      - Set debugger breakpoint
      - Set initial run-time configuration values.

    Exit Criteria:
      - None.
    """

    def __init__(self, name="start_of_simulation"):
        UVMBottomupPhase.__init__(self, name)

    def exec_func(self, comp, phase):
        comp.start_of_simulation_phase(phase)

    m_inst = None  # uvm_start_of_simulation_phase
    type_name = "uvm_start_of_simulation_phase"

    @classmethod
    def get(cls):
        """
        Function: get
        Returns the singleton phase handle
        Returns:
        """
        if UVMStartofSimulationPhase.m_inst is None:
            UVMStartofSimulationPhase.m_inst = UVMStartofSimulationPhase()
        return UVMStartofSimulationPhase.m_inst

    def get_type_name(self):
        return UVMStartofSimulationPhase.type_name
    #endclass



class UVMRunPhase(UVMTaskPhase):
    """
    Stimulate the DUT.

    This `UVMTaskPhase` calls the
    `UVMComponent.run_phase` virtual method. This phase runs in
    parallel to the runtime phases, `UVMPreResetPhase` through
    `UVMPostShutdownPhase`. All components in the testbench
    are synchronized with respect to the run phase regardless of
    the phase domain they belong to.

    Upon Entry:
    - Indicates that power has been applied.
    - There should not have been any active clock edges before entry
      into this phase (e.g. x->1 transitions via initial blocks).
    - Current simulation time is still equal to 0
      but some "delta cycles" may have occurred.

    Typical Uses:
    - Components implement behavior that is exhibited for the entire
      run-time, across the various run-time phases.
    - Backward compatibility with OVM.

    Exit Criteria:
    - The DUT no longer needs to be simulated, and
    - The <uvm_post_shutdown_phase> is ready to end

    The run phase terminates in one of two ways.

    1. All run_phase objections are dropped:

      When all objections on the run_phase objection have been dropped,
      the phase ends and all of its threads are killed.
      If no component raises a run_phase objection immediately upon
      entering the phase, the phase ends immediately.


    2. Timeout:

      The phase ends if the timeout expires before all objections are dropped.
      By default, the timeout is set to 9200 seconds.
      You may override this via <uvm_root::set_timeout>.

      If a timeout occurs in your simulation, or if simulation never
      ends despite completion of your test stimulus, then it usually indicates
      that a component continues to object to the end of a phase.
    """


    async def exec_task(self, comp, phase):
        uvm_debug(self, 'exec_task', comp.get_name() + ' yielding comp.run_phase()')
        # tpoikela, modification  of original to allow handle for proc
        #yield comp.run_phase(phase)
        comp.m_run_process = cocotb.fork(comp.run_phase(phase))
        await comp.m_run_process

        #parent = comp.get_parent()
        #if parent is not None and parent.get_name() == '__top__':
        #    for i in range(10):
        #        yield Timer(0, "NS")
        #else:
        #    if parent is not None:
        #            comp.get_name())
        uvm_debug(self, 'exec_task', comp.get_name() + ' returned from comp.run_phase()')

    m_inst = None  # static uvm_run_phase
    type_name = "uvm_run_phase"

    @classmethod
    def get(cls):
        """
          Function: get
          Returns the singleton phase handle
        Returns:
        """
        if UVMRunPhase.m_inst is None:
            UVMRunPhase.m_inst = UVMRunPhase()
        return UVMRunPhase.m_inst

    def __init__(self, name="run"):
        UVMTaskPhase.__init__(self, name)

    def get_type_name(self):
        return UVMRunPhase.type_name


class UVMExtractPhase(UVMBottomupPhase):
    """
    Extract data from different points of the verification environment.

    `UVMBottomupPhase` that calls the
    `UVMComponent.extract_phase` method.

    Upon Entry:
    - The DUT no longer needs to be simulated.
    - Simulation time will no longer advance.

    Typical Uses:
    - Extract any remaining data and final state information
      from scoreboard and testbench components
    - Probe the DUT (via zero-time hierarchical references
      and/or backdoor accesses) for final state information.
    - Compute statistics and summaries.
    - Display final state information
    - Close files.

    Exit Criteria:
    - All data has been collected and summarized.
    """

    def exec_func(self, comp, phase):
        comp.extract_phase(phase)

    m_inst = None  # uvm_extract_phase
    type_name = "uvm_extract_phase"

    @classmethod
    def get(cls):
        """
        Function: get
        Returns the singleton phase handle
        Returns:
        """
        if UVMExtractPhase.m_inst is None:
            UVMExtractPhase.m_inst = UVMExtractPhase()
        return UVMExtractPhase.m_inst

    def __init__(self, name="extract"):
        UVMBottomupPhase.__init__(self, name)

    def get_type_name(self):
        return UVMExtractPhase.type_name


class UVMCheckPhase(UVMBottomupPhase):
    """
    Check for any unexpected conditions in the verification environment.

    `UVMBottomupPhase` that calls the
    `UVMComponent.check_phase` method.

    Upon Entry:
    - All data has been collected.

    Typical Uses:
    - Check that no unaccounted-for data remain.

    Exit Criteria:
    - Test is known to have passed or failed.
    """


    def exec_func(self, comp, phase):
        comp.check_phase(phase)

    m_inst = None  # uvm_check_phase

    type_name = "uvm_check_phase"

    @classmethod
    def get(cls):
        """     Function: get
        Returns the singleton phase handle
        Returns:
        """
        if UVMCheckPhase.m_inst is None:
            UVMCheckPhase.m_inst = UVMCheckPhase()
        return UVMCheckPhase.m_inst

    def __init__(self, name="check"):
        UVMBottomupPhase.__init__(self, name)

    def get_type_name(self):
        return UVMCheckPhase.type_name


class UVMReportPhase(UVMBottomupPhase):
    """
    Report results of the test.

    `UVMBottomupPhase` that calls the
    `UVMComponent.report_phase` method.

    Upon Entry:
    - Test is known to have passed or failed.

    Typical Uses:
    - Report test results.
    - Write results to file.

    Exit Criteria:
    - End of test.
    """

    def exec_func(self, comp, phase):
        comp.report_phase(phase)

    m_inst = None  # uvm_report_phase

    type_name = "uvm_report_phase"

    @classmethod
    def get(cls):
        """     Function: get
        Returns the singleton phase handle
        Returns:
        """
        if UVMReportPhase.m_inst is None:
            UVMReportPhase.m_inst = UVMReportPhase()
        return UVMReportPhase.m_inst

    def __init__(self, name="report"):
        UVMBottomupPhase.__init__(self, name)

    def get_type_name(self):
        return UVMReportPhase.type_name


class UVMFinalPhase(UVMBottomupPhase):
    """
    Tie up loose ends.

    `UVMTopdownPhase` that calls the
    `UVMComponent.final_phase` method.

    Upon Entry:
    - All test-related activity has completed.

    Typical Uses:
    - Close files.
    - Terminate co-simulation engines.

    Exit Criteria:
    - Ready to exit simulator.
    """

    def exec_func(self, comp, phase):
        comp.final_phase(phase)

    m_inst = None  # static uvm_final_phase

    type_name = "uvm_final_phase"

    @classmethod
    def get(cls):
        """     Function: get
        Returns the singleton phase handle
        Returns:
        """
        if UVMFinalPhase.m_inst is None:
            UVMFinalPhase.m_inst = UVMFinalPhase()
        return UVMFinalPhase.m_inst

    def __init__(self, name="final"):
        UVMBottomupPhase.__init__(self, name)

    def get_type_name(self):
        return UVMFinalPhase.type_name



#//----------------------------------------------------------------------
#// Copyright 2010-2011 Mentor Graphics Corporation
#// Copyright 2010 Synopsys, Inc.
#// Copyright 2010-2018 Cadence Design Systems, Inc.
#// Copyright 2015 NVIDIA Corporation
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.base.uvm_port_base import UVMPortBase
from uvm.tlm1.uvm_tlm_imps import UVM_EXPORT_COMMON
from uvm.tlm2.uvm_tlm2_defines import UVM_TLM_B_MASK, UVM_TLM_NB_FW_MASK,\
    UVM_TLM_NB_BW_MASK
from uvm.tlm2.uvm_tlm2_imps import UVM_TLM_B_TRANSPORT_IMP,\
    UVM_TLM_NB_TRANSPORT_FW_IMP, UVM_TLM_NB_TRANSPORT_BW_IMP
#
#//----------------------------------------------------------------------
#// Title -- NODOCS -- TLM2 Export Classes
#//
#// This section defines the export classes for connecting TLM2
#// interfaces.
#//----------------------------------------------------------------------
#
#
#// Class -- NODOCS -- uvm_tlm_b_transport_export
#//
#// Blocking transport export class.

class UVMTLMBTransportExport(UVMPortBase):
    pass

UVMTLMBTransportExport = UVM_EXPORT_COMMON(UVMTLMBTransportExport, UVM_TLM_B_MASK, "uvm_tlm_b_transport_export")
UVM_TLM_B_TRANSPORT_IMP('m_if', UVMTLMBTransportExport)



#// Class: uvm_tlm_nb_transport_fw_export
#//
#// Non-blocking forward transport export class

class UVMTLMNbTransportFwExport(UVMPortBase):
    pass

UVMTLMNbTransportFwExport = UVM_EXPORT_COMMON(UVMTLMNbTransportFwExport, UVM_TLM_NB_FW_MASK,
        "uvm_tlm_nb_transport_fw_export")
UVM_TLM_NB_TRANSPORT_FW_IMP('m_if', UVMTLMNbTransportFwExport)



#// Class: uvm_tlm_nb_transport_bw_export
#//
#// Non-blocking backward transport export class

class UVMTLMNbTransportBwExport(UVMPortBase):
    pass

UVMTLMNbTransportBwExport = UVM_EXPORT_COMMON(UVMTLMNbTransportBwExport, UVM_TLM_NB_BW_MASK,
        "uvm_tlm_nb_transport_bw_export")
UVM_TLM_NB_TRANSPORT_BW_IMP('m_if', UVMTLMNbTransportBwExport)




#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc. 
#//   Copyright 2010 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------
#
#
#//------------------------------------------------------------------------------
#//
#// CLASS: uvm_push_sequencer #(REQ,RSP)
#//
#//------------------------------------------------------------------------------
#
#class uvm_push_sequencer #(type REQ=uvm_sequence_item, RSP=REQ)
#                                   extends uvm_sequencer_param_base #(REQ, RSP);
#
#  typedef uvm_push_sequencer #( REQ , RSP) this_type;
#
#  // Port: req_port
#  //
#  // The push sequencer requires access to a blocking put interface.
#  // A continuous stream of sequence items are sent out this port, based on
#  // the list of available sequences loaded into this sequencer.
#  //
#  uvm_blocking_put_port #(REQ) req_port;
#
#
#  // Function: new
#  //
#  // Standard component constructor that creates an instance of this class
#  // using the given ~name~ and ~parent~, if any.
#  //
#  function new (string name, uvm_component parent=null);
#    super.new(name, parent);
#    req_port = new ("req_port", this);
#  endfunction 
#
#
#  // Task: run_phase
#  //
#  // The push sequencer continuously selects from its list of available
#  // sequences and sends the next item from the selected sequence out its
#  // <req_port> using req_port.put(item). Typically, the req_port would be
#  // connected to the req_export on an instance of a
#  // <uvm_push_driver #(REQ,RSP)>, which would be responsible for
#  // executing the item.
#  //
#  task run_phase(uvm_phase phase);
#    REQ t;
#    int selected_sequence;
#
#    fork
#      super.run_phase(phase);
#      forever
#        begin
#          m_select_sequence();
#          m_req_fifo.get(t);
#          req_port.put(t);
#          m_wait_for_item_sequence_id = t.get_sequence_id();
#          m_wait_for_item_transaction_id = t.get_transaction_id();
#        end
#    join
#  endtask
#
#  protected virtual function int  m_find_number_driver_connections();
#    return req_port.size();
#  endfunction
#
#endclass





import unittest

from uvm.base.uvm_report_handler import UVMReportHandler
from uvm.base.uvm_object_globals import *


class TestUVMReportHandler(unittest.TestCase):

    def test_handler(self):
        rh = UVMReportHandler("handler")
        fhandle = rh.get_severity_id_file(UVM_ERROR, "")
        self.assertEqual(fhandle, rh.default_file_handle)
        verb = rh.get_verbosity_level()
        self.assertEqual(verb, rh.m_max_verbosity_level)
        rh.report(UVM_ERROR, "xxx", "id0", "Message")


    def test_action_overrides(self):
        rh = UVMReportHandler("handler")
        rh.set_severity_action(UVM_ERROR, UVM_LOG)
        rh.set_severity_id_action(UVM_ERROR, "XYZ", UVM_COUNT)

        verb = rh.get_action(UVM_ERROR, "ABC")
        self.assertEqual(verb, UVM_LOG)
        verb = rh.get_action(UVM_ERROR, "XYZ")
        self.assertEqual(verb, UVM_COUNT)


if __name__ == '__main__':
    unittest.main()
